import copy

from uptrain.core.lib.helper_funcs import extract_data_points_from_batch
from uptrain.core.classes.statistics import AbstractStatistic
from uptrain.core.classes.distances import DistanceResolver
from uptrain.core.classes.measurables import MeasurableResolver
from uptrain.constants import Statistic


class Distance(AbstractStatistic):
    dashboard_name = "distance"
    statistic_type = Statistic.DISTANCE

    def base_init(self, fw, check):
        self.aggregate_measurable = MeasurableResolver(check["aggregate_args"]).resolve(
            fw
        )
        self.count_measurable = MeasurableResolver(check["count_args"]).resolve(
            fw
        )
        self.item_counts = {}
        self.feats_dictn = {}
        self.reference = check["reference"]
        self.distance_types = check["distance_types"]
        self.dist_classes = [DistanceResolver().resolve(x) for x in self.distance_types]

    def base_check(self, inputs, outputs=None, gts=None, extra_args={}):
        vals = self.measurable.compute_and_log(
            inputs, outputs, gts=gts, extra=extra_args
        )
        aggregate_ids = self.aggregate_measurable.compute_and_log(
            inputs, outputs, gts=gts, extra=extra_args
        )
        counts = self.count_measurable.compute_and_log(
            inputs, outputs, gts=gts, extra=extra_args
        )
        all_models = [x.compute_and_log(
            inputs, outputs, gts=gts, extra=extra_args
        ) for x in self.model_measurables]
        all_features = [x.compute_and_log(
            inputs, outputs, gts=gts, extra=extra_args
        ) for x in self.feature_measurables]

        for idx in range(len(aggregate_ids)):
            is_model_invalid = sum([all_models[jdx][idx] not in self.allowed_model_values[jdx] for jdx in range(len(self.allowed_model_values))])
            if is_model_invalid:
                continue

            this_val = extract_data_points_from_batch(vals, [idx])

            if aggregate_ids[idx] not in self.item_counts:
                self.item_counts.update({aggregate_ids[idx]: 0})
                self.feats_dictn.update({aggregate_ids[idx]: this_val})

            this_item_count_prev = self.item_counts[aggregate_ids[idx]]
            if this_item_count_prev == counts[idx]:
                continue
            self.item_counts[aggregate_ids[idx]] = counts[idx]
            this_item_count = self.item_counts[aggregate_ids[idx]]

            this_distances = dict(
                zip(
                    self.distance_types,
                    [
                        x.compute_distance(
                            this_val, self.feats_dictn[aggregate_ids[idx]]
                        )
                        for x in self.dist_classes
                    ],
                )
            )
            if self.reference == "running_diff":
                self.feats_dictn[aggregate_ids[idx]] = this_val
            models = dict(zip(['model_' + x for x in self.model_names], [all_models[jdx][idx] for jdx in range(len(self.model_names))]))
            features = dict(zip(['feature_' + x for x in self.feature_names], [all_features[jdx][idx] for jdx in range(len(self.feature_names))]))
            for distance_type in self.distance_types:
                plot_name = (
                    distance_type
                    + "_"
                    + str(self.reference)
                )
                self.log_handler.add_scalars(
                    self.dashboard_name + "_" + plot_name,
                    {'y_' + distance_type: this_distances[distance_type][0]},
                    this_item_count,
                    self.dashboard_name,
                    features = features,
                    models = models,
                    file_name = str(aggregate_ids[idx])
                )


from uptrain.core.classes.measurables import (
    Measurable,
    FeatureMeasurable,
    ConditionMeasurable,
    CustomMeasurable,
    AccuracyMeasurable,
    MAEMeasurable,
    MAPEMeasurable,
    ScalarFromEmbeddingMeasurable,
    DistanceMeasurable,
    RecHitRateMeasurable,
)
from uptrain.constants import MeasurableType


class MeasurableResolver:
    def __init__(self, args) -> None:
        super().__init__()
        self._args = args

    def has_valid_resolve_args(self):
        if self._args is None:
            return False

        if len(self._args) == 0:
            return False

        return True

    def resolve(self, framework) -> Measurable:
        if not self.has_valid_resolve_args():
            return None

        resolve_args = self._args
        measurable_type = resolve_args["type"]
        if measurable_type == MeasurableType.INPUT_FEATURE:
            return FeatureMeasurable(framework, resolve_args["feature_name"], "inputs")
        elif measurable_type == MeasurableType.PREDICTION:
            return FeatureMeasurable(framework, resolve_args["feature_name"], "outputs")
        elif measurable_type == MeasurableType.CUSTOM:
            return CustomMeasurable(framework, resolve_args)
        elif measurable_type == MeasurableType.ACCURACY:
            return AccuracyMeasurable(framework)
        elif measurable_type == MeasurableType.MAE:
            return MAEMeasurable(framework)
        elif measurable_type == MeasurableType.MAPE:
            return MAPEMeasurable(framework)
        elif measurable_type == MeasurableType.CONDITION_ON_INPUT:
            return ConditionMeasurable(
                framework,
                FeatureMeasurable(framework, resolve_args["feature_name"], "inputs"),
                resolve_args["condition_args"],
            )
        elif measurable_type == MeasurableType.CONDITION_ON_PREDICTION:
            return ConditionMeasurable(
                framework,
                FeatureMeasurable(framework, resolve_args["feature_name"], "outputs"),
                resolve_args["condition_args"],
            )
        elif measurable_type == MeasurableType.SCALAR_FROM_EMBEDDING:
            return ScalarFromEmbeddingMeasurable(
                framework, resolve_args["idx"], resolve_args["extract_from"]
            )
        elif measurable_type == MeasurableType.DISTANCE:
            return DistanceMeasurable(
                framework,
                resolve_args["base"],
                resolve_args["reference"],
                resolve_args["distance_types"],
            )
        elif measurable_type == MeasurableType.REC_HIT_RATE:
            return RecHitRateMeasurable(framework)
        else:
            raise Exception("Resolver not defined")



import numpy as np

from uptrain.core.classes.measurables import Measurable


class RecHitRateMeasurable(Measurable):
    def __init__(self, framework) -> None:
        super().__init__(framework)

    def _compute(self, inputs=None, outputs=None, gts=None, extra=None) -> any:
        acc = [1 if gt in outputs[i] else 0 for i,gt in enumerate(gts)]
        return np.array(acc)

    def col_name(self):
        return "Hit-Rate"


import shap
import pickle
from uptrain.core.classes.visuals import AbstractVisual
from uptrain.constants import Visual
import os

class Shap(AbstractVisual):
    visual_type = Visual.SHAP
    dashboard_name = "SHAP_explanation"

    def base_init(self, fw, check):
        self.model = check["model"]
        self.explainer_created = False

        filename = 'explainer.pkl'
        self.fileloc = os.path.join(fw.log_handler.st_log_folder, filename)

        fw.log_handler.add_st_metadata(
            {'path_all_data': fw.path_all_data,
            'shap_num_points': check.get("shap_num_points", 1000),
            'path_shap_file': self.fileloc})

    def base_check(self, inputs, outputs, gts=None, extra_args={}):
        if not self.explainer_created:
            explainer = shap.Explainer(self.model)
            file = open(self.fileloc, 'wb')
            pickle.dump(explainer, file)
            file.close()
            self.explainer_created = True




from uptrain.core.classes.measurables import (
    Measurable,
    InputFeatureMeasurable,
    OutputFeatureMeasurable,
)


class FeatureMeasurable(Measurable):
    def __init__(self, framework, feature_name, dictn_type) -> None:
        super().__init__(framework)
        self.feature_name = feature_name
        self.dictn_type = dictn_type
        if self.dictn_type == "inputs":
            self.helper = InputFeatureMeasurable(framework, feature_name)
        elif self.dictn_type == "outputs":
            self.helper = OutputFeatureMeasurable(framework, feature_name)
        else:
            raise Exception(
                "Helper Measurable not defined for dictionary type %s" % self.dictn_type
            )

    def _compute(self, inputs=None, outputs=None, gts=None, extra=None) -> any:
        return self.helper._compute(
            inputs=inputs, outputs=outputs, gts=gts, extra=extra
        )

    def col_name(self):
        return self.helper.col_name()

    # TODO: Decommission and find a generic way
    def extract_val_from_training_data(self, x):
        return self.helper.extract_val_from_training_data(x)



import numpy as np

from river import drift

from uptrain.core.classes.monitors import AbstractMonitor
from uptrain.constants import DataDriftAlgo, MeasurableType
from uptrain.core.classes.measurables import MeasurableResolver
from uptrain.constants import Monitor


class ConceptDrift(AbstractMonitor):
    dashboard_name = "concept_drift_acc"
    monitor_type = Monitor.CONCEPT_DRIFT

    def base_init(self, fw, check):
        if check.get("measurable_args", None):
            self.measurable = MeasurableResolver(check["measurable_args"]).resolve(fw)
        else:
            self.measurable = MeasurableResolver({"type": MeasurableType.ACCURACY}).resolve(fw)

        self.acc_arr = []
        self.avg_acc = 0
        self.drift_alerted = False
        self.algorithm = check["algorithm"]
        self.counter = 0

        if self.algorithm == DataDriftAlgo.DDM:
            warm_start = check.get("warm_start", 500)
            warn_threshold = check.get("warn_threshold", 2.0)
            alarm_threshold = check.get("alarm_threshold", 3.0)
            self.algo = drift.DDM(warm_start, warn_threshold, alarm_threshold)
        elif self.algorithm == DataDriftAlgo.ADWIN:
            delta = check.get("delta", 0.002)
            clock = check.get("clock", 32)
            max_buckets = check.get("max_buckets", 5)
            min_window_length = check.get("min_window_length", 5)
            grace_period = check.get("grace_period", 5)
            self.algo = drift.ADWIN(delta, clock, max_buckets, min_window_length, grace_period)
        else:
            raise Exception("Data drift algo type not supported")

    def need_ground_truth(self):
        return True

    def base_check(self, inputs, outputs, gts=None, extra_args={}):
        batch_acc = self.measurable.compute_and_log(inputs, outputs, gts, extra_args)
        batch_acc = self._preprocess(batch_acc)

        for acc in batch_acc:
            alert = None
            self.algo.update(acc)

            if self.algo.drift_detected and not self.drift_alerted:
                alert = f"Drift detected with {self.algorithm} at time: {self.counter}"
                print(alert)
                self.drift_alerted = True

            self.acc_arr.append(acc)
            self.avg_acc = (self.avg_acc * (len(self.acc_arr) - 1) + acc) / len(
                self.acc_arr
            )
            self.log_handler.add_scalars(
                "avg_accuracy",
                {"y_avg_accuracy": self.avg_acc},
                len(self.acc_arr),
                self.dashboard_name,
            )
            self.counter += 1

            if isinstance(alert, str):
                self.log_handler.add_alert(
                    "Model Performance Degradation Alert 🚨", alert, self.dashboard_name
                )
    
    def _preprocess (self, batch):
        if self.algorithm == DataDriftAlgo.DDM:
            return np.array([0 if x else 1 for x in batch])
        else:
            return np.array(batch)



import numpy as np

from uptrain.core.classes.measurables import Measurable
from uptrain.core.classes.signals import SignalManager


class CustomMeasurable(Measurable):
    def __init__(self, framework, custom_args={}) -> None:
        super().__init__(framework)
        self._args = custom_args
        self.signal_manager = SignalManager()
        self.signal_manager.add_signal_formulae(self._args["signal_formulae"])

    def _compute(self, inputs=None, outputs=None, gts=None, extra=None) -> any:
        val = self.signal_manager.evaluate_signal(
            inputs, outputs, gts=gts, extra_args=extra
        )
        return val

    def col_name(self):
        return str(self.signal_manager)

    # TODO: Decommission and find a generic way
    def extract_val_from_training_data(self, x):
        fake_inputs = {}
        for key, val in x.items():
            fake_inputs.update({key: np.array([val])})
        return self.signal_manager.evaluate_signal(fake_inputs, None)



import numpy as np
import copy

from uptrain.core.lib.helper_funcs import extract_data_points_from_batch
from uptrain.core.classes.distances import DistanceResolver
from uptrain.core.classes.statistics import AbstractStatistic
from uptrain.core.classes.measurables import MeasurableResolver
from uptrain.constants import Statistic
# from uptrain.core.classes.algorithms import Clustering
# from uptrain.core.lib.algorithms import estimate_earth_moving_cost

class Convergence(AbstractStatistic):
    dashboard_name = "convergence_stats"
    statistic_type = Statistic.CONVERGENCE_STATS

    def base_init(self, fw, check):
        self.aggregate_measurable = MeasurableResolver(check["aggregate_args"]).resolve(
            fw
        )
        self.count_measurable = MeasurableResolver(check["count_args"]).resolve(
            fw
        )
        self.item_counts = {}
        self.first_count_checkpoint = {}
        self.reference = check["reference"]
        self.distance_types = check["distance_types"]

        self.count_checkpoints = np.unique(np.array([0] + check["count_checkpoints"]))

        self.feats_dictn = dict(zip(list(self.count_checkpoints), [{} for x in list(self.count_checkpoints)]))
        # ex: {0: {agg_id_0: val_0, ..}, 200: {}, 500: {}, 1000: {}, 5000: {}, 20000: {}}

        self.distances_dictn = dict(zip(list(self.count_checkpoints), [dict(zip(self.distance_types, [[] for x in list(self.distance_types)])) for y in list(self.count_checkpoints)]))
        #ex: {0: {'cosine_distance': [d1, d2, ..]}, 200: {'cosine_distance': []}, 500: {'cosine_distance': []}, 1000: {'cosine_distance': []}, 5000: {'cosine_distance': []}, 20000: {'cosine_distance': []}}

        self.dist_classes = [DistanceResolver().resolve(x) for x in self.distance_types]
        self.total_count = 0
        self.prev_calc_at = 0

    def base_check(self, inputs, outputs, gts=None, extra_args={}):
        vals = self.measurable.compute_and_log(
            inputs, outputs, gts=gts, extra=extra_args
        )
        aggregate_ids = self.aggregate_measurable.compute_and_log(
            inputs, outputs, gts=gts, extra=extra_args
        )
        counts = self.count_measurable.compute_and_log(
            inputs, outputs, gts=gts, extra=extra_args
        )
        all_models = [x.compute_and_log(
            inputs, outputs, gts=gts, extra=extra_args
        ) for x in self.model_measurables]
        all_features = [x.compute_and_log(
            inputs, outputs, gts=gts, extra=extra_args
        ) for x in self.feature_measurables]
        self.total_count += len(extra_args['id'])

        models = dict(zip(['model_' + x for x in self.model_names], [self.allowed_model_values[jdx][0] for jdx in range(len(self.model_names))]))

        for idx in range(len(aggregate_ids)):
            is_model_invalid = sum([all_models[jdx][idx] not in self.allowed_model_values[jdx] for jdx in range(len(self.allowed_model_values))])
            if is_model_invalid:
                continue

            if aggregate_ids[idx] not in self.item_counts:
                self.item_counts.update({aggregate_ids[idx]: 0})

            this_item_count_prev = self.item_counts[aggregate_ids[idx]]
            if this_item_count_prev >= counts[idx]:
                continue
            self.item_counts[aggregate_ids[idx]] = counts[idx]
            this_item_count = self.item_counts[aggregate_ids[idx]]

            crossed_checkpoint_arr = (np.logical_xor((self.count_checkpoints > this_item_count), (self.count_checkpoints > this_item_count_prev)))
            has_crossed_checkpoint = sum(crossed_checkpoint_arr)

            if has_crossed_checkpoint:
                crossed_checkpoint = self.count_checkpoints[np.where(crossed_checkpoint_arr == True)[0][-1]]

                if aggregate_ids[idx] not in self.first_count_checkpoint:
                    self.first_count_checkpoint.update({aggregate_ids[idx]: crossed_checkpoint})

                this_val = extract_data_points_from_batch(vals, [idx])

                self.feats_dictn[crossed_checkpoint].update({aggregate_ids[idx]: this_val})
                if crossed_checkpoint > self.first_count_checkpoint[aggregate_ids[idx]]:
                    if self.reference == "running_diff":
                        this_count_idx = np.where(self.count_checkpoints == crossed_checkpoint)[0][0]
                        prev_count_idx = max(0, this_count_idx - 1)
                        found_ref = False
                        while not found_ref:
                            ref_item_count = self.count_checkpoints[prev_count_idx]
                            if aggregate_ids[idx] in self.feats_dictn[ref_item_count]:
                                break
                            prev_count_idx = max(0, prev_count_idx - 1)
                    else:
                        ref_item_count = self.first_count_checkpoint[aggregate_ids[idx]]
                    this_distances = dict(
                        zip(
                            self.distance_types,
                            [
                                x.compute_distance(
                                    this_val,
                                    self.feats_dictn[ref_item_count][
                                        aggregate_ids[idx]
                                    ],
                                )
                                for x in self.dist_classes
                            ],
                        )
                    )
                    if self.reference == "running_diff":
                        del self.feats_dictn[ref_item_count][aggregate_ids[idx]]
                    else:
                        del self.feats_dictn[crossed_checkpoint][aggregate_ids[idx]]

                    features = dict(zip(['feature_' + x for x in self.feature_names], [all_features[jdx][idx] for jdx in range(len(self.feature_names))]))

                    for distance_key in list(this_distances.keys()):
                        plot_name = (
                            distance_key
                            + " "
                            + str(self.reference)
                        )
                        this_data = list(np.reshape(np.array(this_distances[distance_key]),-1))

                        self.distances_dictn[crossed_checkpoint][distance_key].extend(
                            this_data
                        )

                        self.log_handler.add_histogram(
                            plot_name,
                            this_data,
                            self.dashboard_name,
                            models = [models]*len(this_data),
                            features = [features] * len(this_data),
                            file_name = str(crossed_checkpoint)
                        )


        if ((self.total_count - self.prev_calc_at) > 50000):
            self.prev_calc_at = self.total_count
            for count in list(self.distances_dictn.keys()):
                if (count > 0):
                    for distance_type in self.distance_types:
                        plot_name = (
                            distance_type
                            + " "
                            + str(self.reference)
                        )
                        this_data = np.reshape(
                                np.array(self.distances_dictn[count][distance_type]), -1
                            )

                        if len(this_data) > 5:
                            self.log_handler.add_scalars(
                                plot_name + "_mean",
                                {'y_mean': np.mean(this_data)},
                                count,
                                self.dashboard_name,
                                models = models,
                                features = {"tagGenre": "All"},
                                file_name = str("count"),
                                update_val = True
                            )

                            # next_count_idx = np.where(self.count_checkpoints == (count))[0][0] + 1
                            # if next_count_idx < len(self.count_checkpoints):
                            #     next_data = np.reshape(
                            #         np.array(self.distances_dictn[self.count_checkpoints[next_count_idx]][distance_type]), -1
                            #     )
                            #     if len(next_data) > 5:
                            #         clustering_helper = Clustering({"num_buckets": 2, "is_embedding": False})
                            #         this_data = np.expand_dims(np.array(this_data), axis=(1))
                            #         next_data = np.expand_dims(np.array(next_data), axis=(1,2))
                            #         this_count_clustering_res = clustering_helper.cluster_data(this_data)
                            #         next_count_clustering_res = clustering_helper.infer_cluster_assignment(next_data)
                            #         emd_cost = estimate_earth_moving_cost(np.reshape(next_count_clustering_res[1]/next_data.shape[0],-1), np.reshape(clustering_helper.dist[0],-1), clustering_helper.clusters[0])
                            #         self.log_handler.add_scalars(
                            #             plot_name + "_emd",
                            #             {'y_distance': emd_cost},
                            #             count,
                            #             # self.total_count,                     
                            #             self.dashboard_name,
                            #             models = models,
                            #             features = {"tagGenre": "All"},
                            #             file_name = str("count"),
                            #             update_val = True
                            #         )


from abc import ABC


class AbstractMeasurable(ABC):
    def __init__(self) -> None:
        super().__init__()

    def _compute(self, inputs=None, outputs=None, gts=None, extra=None) -> any:
        raise Exception("Should be defined for each individual class")



import numpy as np

from uptrain.core.classes.monitors import AbstractMonitor
from uptrain.core.classes.algorithms import PopularityBias
from uptrain.constants import BiasAlgo, Monitor


class ModelBias(AbstractMonitor):
    dashboard_name = "popularity_bias"
    monitor_type = Monitor.POPULARITY_BIAS

    def base_init(self, fw, check):
        self.acc_arr = []
        if check["algorithm"] == BiasAlgo.POPULARITY_BIAS:
            sessions = check.get("sessions", None)
            self.algo = PopularityBias(sessions)
        else:
            raise Exception("Model bias type not supported")

    def need_ground_truth(self):
        return False

    def base_check(self, inputs, outputs, gts=None, extra_args={}):
        for y_pred in outputs:
            self.algo.add_prediction(y_pred)
            pop_arr = self.algo.all_popularity
            self.log_handler.add_histogram(
                "popularity_bias", pop_arr, self.dashboard_name
            )



class ModelHandler:
    """
    Handler class to help with all the model training related functionalities

    ...
    Attributes
    ----------
    training_func : func
        Used to train the model
    inference_func: func
        Used to evaluate the model

    Methods
    -------
    set_training_func(func):
        Attach the model training func
    set_inference_func(func):
        Attach the model evaluation func
    retrain():
        Retrains the model using the provided training dataset. Saves under new version
    compare_new_model(testing_dataset):
        Compares the new model against the old model on the given testing dataset
    """

    def set_training_func(self, func):
        """Attach model training pipeline"""

        self.training_func = func

    def retrain(self, training_file, new_model_version):
        """Calls the training func on the given training_file
        New model is saved under the name: 'version_' + provided new_model_version

        Parameters
        ----------
        training_file : str
            Path to the file used to train the model
        new_model_version : int or str
            New model is saved under the name: 'version_' + provided new_model_version
        """

        self.training_func(training_file, "version_" + str(new_model_version))

    def set_inference_func(self, func):
        """Attach model evaluation pipeline"""

        self.inference_func = func

    def compare_new_model(
        self, testing_dataset, new_model_version, old_training_dataset, selected_count
    ):
        """Compares new model against the old model on testing dataset

        Parameters
        ----------
        testing_dataset : str
            Path to the file used to test the model
        new_model_version : int or str
            Old model is trained and saved under the name: 'version_' + provided new_model_version - 1
        old_training_dataset : str
            Training file for the old model
        selected_count : int
            Number of data-points selected by the framework to add to the retraining dataset
        """

        new_model_name = "version_" + str(new_model_version)
        old_model_name = "version_" + str(new_model_version - 1)

        # Train the old model
        # TODO: No need for this, we can just use the deployed old model
        self.training_func(old_training_dataset, old_model_name)

        # Compute and print accuracies of the old and new model
        old_model_accuracy = self.inference_func(testing_dataset, old_model_name)
        new_model_accuracy = self.inference_func(testing_dataset, new_model_name)

        print("---------------------------------------------")
        print("---------------------------------------------")
        print("Old model accuracy: ", old_model_accuracy)
        print(
            "Retrained model accuracy (ie "
            + str(selected_count)
            + " smartly collected data-points added): ",
            new_model_accuracy,
        )
        print("---------------------------------------------")
        print("---------------------------------------------")



import numpy as np

from uptrain.core.classes.monitors import AbstractMonitor
from uptrain.core.classes.measurables import MeasurableResolver
from uptrain.constants import Monitor


class DataIntegrity(AbstractMonitor):
    dashboard_name = "data_integrity"
    monitor_type = Monitor.DATA_INTEGRITY

    def base_init(self, fw, check):
        self.integrity_type = check["integrity_type"]
        self.threshold = check.get("threshold", None)
        self.count = 0
        self.num_issues = 0

    def base_check(self, inputs, outputs, gts=None, extra_args={}):
        signal_value = self.measurable.compute_and_log(
            inputs, outputs, gts=gts, extra=extra_args
        )
        if self.integrity_type == "non_null":
            has_issue = signal_value == None
        elif self.integrity_type == "less_than":
            has_issue = signal_value > self.threshold
        elif self.integrity_type == "greater_than":
            has_issue = signal_value < self.threshold
        self.count += len(signal_value)
        self.num_issues += np.sum(np.array(has_issue))
        plot_name = (
            self.measurable.col_name()
            + " "
            + self.integrity_type
            + " "
            + str(self.threshold)
        )
        self.log_handler.add_scalars(
            self.dashboard_name + "_" + plot_name,
            {"y_" + plot_name: 1 - self.num_issues / self.count},
            self.count,
            self.dashboard_name,
        )

    def need_ground_truth(self):
        return False



from uptrain.core.classes.signals import Signal
from uptrain.constants import ModelSignal


class SignalManager:
    """
    Manager class to help with signal evalutions

    ...
    Attributes
    ----------
    formulae : Signal
        Signal used to identify interesting data-points

    Methods
    -------
    add_signal_formulae(formulae):
        Attach the defined signal formulae to identify interesting data-points
    evaluate_signal(inputs, outputs, extra_args={}):
        Evaluate the defined signal formulae on given inputs and outputs
    """

    def __init__(self):
        """Initialise with default ie pass all signal"""

        self.formulae = Signal(ModelSignal.DEFAULT, is_model_signal=True)

    def add_signal_formulae(self, formulae):
        """Attach the defined signal formulae to identify interesting data-points"""

        self.formulae = formulae

    def evaluate_signal(self, inputs, outputs, gts=None, extra_args={}):
        """Evaluate the defined signal formulae on given inputs and outputs"""
        return self.formulae.evaluate(inputs, outputs, gts=None, extra_args=extra_args)

    def __str__(self) -> str:
        return str(self.formulae)



from uptrain.core.classes.distances import (
    CosineDistance,
    L2Distance,
    NormRatio,
)


class DistanceResolver:
    def __init__(self) -> None:
        pass

    def resolve(self, distance_type):
        if distance_type == "cosine_distance":
            return CosineDistance()
        elif distance_type == "l2_distance":
            return L2Distance()
        elif distance_type == "norm_ratio":
            return NormRatio()
        else:
            raise Exception("distance_type %s is not defined" % distance_type)



"""
This module Implements storage backends for logs collected by the Uptrain package. 
For now, they all live on the local filesystem. 

- sqlite
- duckdb? (TODO)
"""

import sqlite3
from typing import Any, Sequence, Dict, List, Union, Callable

import numpy as np
import pandas as pd


def make_encoder(x: Any) -> Callable:
    """Create an encoder to encode all values with the same type as the one passed."""
    if isinstance(x, (int, float, str)):
        return lambda x: x
    elif isinstance(x, (np.integer, np.floating, np.bool_)):
        return lambda x: x.item()
    elif isinstance(x, np.ndarray):
        return lambda x: str(x.tolist())
    else:
        return lambda x: str(x)


def make_sqlite_type(x: Any) -> str:
    """Produce the sqlite type to map to for the given value."""
    if isinstance(x, (int, bool, np.integer, np.bool_)):
        return "INTEGER"
    elif isinstance(x, (float, np.floating)):
        return "REAL"
    elif isinstance(x, str):
        return "TEXT"
    else:
        return "BLOB"


class SqliteStore:
    """A sqlite based storage backend for model usage logs."""

    path_db: str
    conn: sqlite3.Connection
    _schema: Union[Dict[str, str], None]
    _encoders: List[Callable]

    def __init__(self, path: str):
        """Initializes a sqlite database at the given path. Removes the existing file if present.

        Args:
            path: path to the sqlite database file.
        """
        self.path_db = path if path == ":memory:" else f"{path}.db"
        self.conn = sqlite3.connect(self.path_db, isolation_level=None)
        with self.conn:
            self.conn.execute("PRAGMA journal_mode=WAL")

        self._schema = None
        self._encoders = []

    def __del__(self):
        self.conn.close()

    def _init_table(self, datapoint: Dict[str, Any]):
        """Infer the schema for the logs table from the first datapoint."""
        self._schema = dict()
        assert (
            "id" in datapoint
        ), "A column named `id` must be present in the logged data."
        for col, val in datapoint.items():
            self._schema[col] = make_sqlite_type(val)
            self._encoders.append(make_encoder(val))

        schema_str = ", ".join(f"{k} {v}" for k, v in self._schema.items())
        with self.conn:
            self.conn.execute(
                f"CREATE TABLE logs ({schema_str}, internal_ts INTEGER DEFAULT (strftime('%s','now')))"
            )

    def log(self, datapoint: Dict[str, Any]):
        """Logs a single datapoint."""
        if self._schema is None:
            self._init_table(datapoint)

        with self.conn:
            self.conn.execute(
                f"INSERT INTO logs ({','.join(datapoint.keys())}) VALUES ({','.join('?' * len(datapoint))})",
                list(self._encoders[i](x) for i, x in enumerate(datapoint.values())),
            )

    def log_many(self, datapoints: Dict[str, Sequence[Any]]):
        """Logs multiple datapoints."""
        if self._schema is None:
            self._init_table({k: v[0] for k, v in datapoints.items()})

        with self.conn:
            self.conn.executemany(
                f"INSERT INTO logs ({','.join(datapoints.keys())}) VALUES ({','.join('?' * len(datapoints))})",
                (
                    [self._encoders[i](x) for i, x in enumerate(ntuple)]
                    for ntuple in zip(*datapoints.values())
                ),
            )

    def fetch_all(self) -> "pd.DataFrame":
        return pd.read_sql_query("SELECT * FROM logs", self.conn)



import numpy as np

from typing import Any, List, Tuple, Union

from uptrain.core.lib.helper_funcs import cluster_and_plot_data


class Clustering:
    def __init__(self, args) -> None:
        """
        Initializes the Clustering object with the specified arguments.

        Parameters
        ----------
        args
            A dictionary containing the following keys:
                - num_buckets: the number of buckets to use when clustering
                - is_embedding: a boolean indicating whether the data is an embedding
                - plot_save_name: a string indicating the name of the plot file to save
                - cluster_plot_func: a function that can be used to plot the clustering results
        """

        self.NUM_BUCKETS = args["num_buckets"]
        self.is_embedding = args["is_embedding"]
        self.plot_save_name = args.get("plot_save_name", "")
        self.cluster_plot_func = args.get("cluster_plot_func", None)
        self.dist = []
        self.dist_counts = []
        self.max_along_axis = []
        self.low_density_regions = []
        self.idxs_closest_to_cluster_centroids = {}

    def cluster_data(self, data: np.ndarray) -> dict:
        """
        Clusters the specified data and returns the results.

        Parameters
        ----------
        data
            The data to be clustered.

        Returns
        -------
        dict
            A dictionary containing the following keys:
                - buckets: the bucket assignments for each data point
                - clusters: the cluster assignments for each data point
                - cluster_vars: the cluster variances for each bucket
                - dist: the distance matrix between each pair of buckets
                - dist_counts: the number of points in each pair of buckets
                - max_along_axis: the maximum value along each axis
                - low_density_regions: the low density regions of the data
                - idxs_closest_to_cluster_centroids: the indices of the data points closest to each cluster centroid
        """

        if self.is_embedding:
            self.bucket_vector(data)
        else:
            buckets = []
            clusters = []
            cluster_vars = []
            for idx in range(data.shape[1]):
                this_inputs = data[:, idx]
                this_buckets, this_clusters, this_cluster_vars = self.bucket_scalar(
                    this_inputs
                )
                buckets.append(this_buckets)
                clusters.append(this_clusters)
                cluster_vars.append(this_cluster_vars)
            self.buckets = np.array(buckets)
            self.clusters = np.array(clusters)
            self.cluster_vars = np.array(cluster_vars)

        self.dist = np.array(self.dist)
        self.dist_counts = np.array(self.dist_counts)

        clustering_results = {
            "buckets": self.buckets,
            "clusters": self.clusters,
            "cluster_vars": self.cluster_vars,
            "dist": self.dist,
            "dist_counts": self.dist_counts,
            "max_along_axis": self.max_along_axis,
            "low_density_regions": self.low_density_regions,
            "idxs_closest_to_cluster_centroids": self.idxs_closest_to_cluster_centroids,
        }

        return clustering_results

    def bucket_scalar(
        self, arr: Union[List[Any], np.ndarray]
    ) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
        """
        Buckets a one-dimensional array.

        Parameters
        ----------
        arr
            The array to be bucketed.

        Returns
        -------
        tuple
            A tuple containing the following elements:
                - buckets: the bucket assignments for each data point
                - clusters: the cluster assignments for each data point
                - cluster_vars: the cluster variances for each bucket
        """

        if isinstance(arr[0], str):
            uniques, counts = np.unique(np.array(arr), return_counts=True)
            buckets = uniques
            self.NUM_BUCKETS = len(buckets)
            clusters = uniques
            cluster_vars = [None] * self.NUM_BUCKETS
            self.ref_dist.append(
                [[counts[x] / len(arr)] for x in range(self.NUM_BUCKETS)]
            )
            self.ref_dist_counts.append([[counts[x]] for x in range(self.NUM_BUCKETS)])
        else:
            sorted_arr = np.sort(arr)
            buckets = []
            clusters = []
            cluster_vars = []
            for idx in range(0, self.NUM_BUCKETS):
                if idx > 0:
                    buckets.append(
                        sorted_arr[int(idx * (len(sorted_arr) - 1) / self.NUM_BUCKETS)]
                    )
                this_bucket_elems = sorted_arr[
                    int((idx) * (len(sorted_arr) - 1) / self.NUM_BUCKETS) : int(
                        (idx + 1) * (len(sorted_arr) - 1) / self.NUM_BUCKETS
                    )
                ]
                gaussian_mean = np.mean(this_bucket_elems)
                gaussian_var = np.var(this_bucket_elems)
                clusters.append([gaussian_mean])
                cluster_vars.append([gaussian_var])

        self.dist.append([[1 / self.NUM_BUCKETS] for x in range(self.NUM_BUCKETS)])
        self.dist_counts.append(
            [[int(len(sorted_arr) / self.NUM_BUCKETS)] for x in range(self.NUM_BUCKETS)]
        )
        return np.array(buckets), np.array(clusters), np.array(cluster_vars)

    def bucket_vector(self, data: np.ndarray) -> None:
        """
        This function takes in an array of vectors and performs bucketing on the data.

        Parameters
        ----------
        data
            A 2D numpy array where each row is a vector to be bucketed.
        """
        abs_data = np.abs(data)
        self.max_along_axis = np.max(abs_data, axis=0)
        data = data / self.max_along_axis

        (
            all_clusters,
            counts,
            cluster_vars,
            density_around_points,
            idxs_closest_to_cluster_centroids,
        ) = cluster_and_plot_data(
            data,
            self.NUM_BUCKETS,
            cluster_plot_func=self.cluster_plot_func,
            plot_save_name=self.plot_save_name,
            normalisation=self.max_along_axis,
        )
        low_density_regions = data[
            np.where(
                density_around_points < np.ceil(len(density_around_points) * 0.002)
            )[0]
        ]

        self.clusters = np.array([all_clusters])
        self.cluster_vars = np.array([cluster_vars])
        self.buckets = self.clusters
        self.low_density_regions = low_density_regions
        self.density_around_points = density_around_points
        self.idxs_closest_to_cluster_centroids = idxs_closest_to_cluster_centroids

        self.dist_counts = np.array([counts])
        self.dist = self.dist_counts / data.shape[0]

    def infer_cluster_assignment(
        self, feats: np.ndarray, prod_dist_counts: np.ndarray = None
    ) -> Tuple[np.ndarray, np.ndarray]:
        """
        Infers the cluster assignments for the given input features.

        Parameters
        ----------
        feats
            The input features for which the cluster assignments are to be inferred.
        prod_dist_counts
            A matrix containing the product of the pairwise distances between each pair
            of buckets and the number of points in each pair of buckets, by default None.

        Returns
        -------
        Tuple[np.ndarray, np.ndarray]
            A tuple containing the inferred cluster assignments for the input features
            and the updated matrix of product of pairwise distances between each pair
            of buckets and number of points in each pair of buckets.
        """

        if prod_dist_counts is None:
            prod_dist_counts = np.zeros((feats.shape[1], self.NUM_BUCKETS))

        if self.is_embedding:
            selected_cluster = np.argmin(
                np.sum(
                    np.abs(self.clusters[0] - feats),
                    axis=tuple(range(2, len(feats.shape))),
                ),
                axis=1,
            )
            for clus in selected_cluster:
                prod_dist_counts[0][clus] += 1
            this_datapoint_cluster = selected_cluster
        else:
            this_datapoint_cluster = []
            for idx in range(feats.shape[2]):
                if isinstance(feats[0, 0, idx], str):
                    try:
                        bucket_idx = np.array(
                            [
                                list(self.buckets[idx]).index(feats[x, 0, idx])
                                for x in range(feats.shape[0])
                            ]
                        )
                    except:
                        # TODO: This logic is not completely tested yet. Contact us if you are facing issues
                        # If given data-point is not present -> add a new bucket
                        temp_buckets = list(self.buckets[idx])
                        num_added = 0

                        for x in range(feats.shape[0]):
                            if feats[x, 0, idx] not in temp_buckets:
                                temp_buckets.append(feats[x, 0, idx])
                                num_added += 1

                        self.buckets[idx] = np.array(temp_buckets)
                        bucket_idx = np.array(
                            [
                                list(self.buckets[idx]).index(feats[x, 0, idx])
                                for x in range(feats.shape[0])
                            ]
                        )
                else:
                    bucket_idx = np.searchsorted(self.buckets[idx], feats[:, :, idx])[
                        :, 0
                    ]
                this_datapoint_cluster.append(bucket_idx)
                for clus in bucket_idx:
                    prod_dist_counts[idx][clus] += 1
            this_datapoint_cluster = np.array(this_datapoint_cluster)
        return this_datapoint_cluster, prod_dist_counts



import numpy as np


class NormRatio:
    def compute_distance(self, base, ref) -> float:
        base = np.array(base)
        ref = np.array(ref)
        if len(base.shape) > 1:
            base_norm = np.linalg.norm(base, axis=1)
        else:
            base_norm = np.abs(base)
        if len(ref.shape) > 1:
            ref_norm = np.linalg.norm(ref, axis=1)
        else:
            ref_norm = np.abs(ref)
        return base_norm / np.maximum(ref_norm, 1e-6)



from .abstract_check import AbstractCheck
from .abstract_monitor import AbstractMonitor
from .accuracy import Accuracy
from .concept_drift import ConceptDrift
from .custom_monitor import CustomMonitor
from .data_drift import DataDrift
from .edge_case import EdgeCase
from .model_bias import ModelBias
from .data_integrity import DataIntegrity


from .abstract import AbstractMeasurable
from .measurable import Measurable
from .input_feature import InputFeatureMeasurable
from .output_feature import OutputFeatureMeasurable
from .feature import FeatureMeasurable
from .condition import ConditionMeasurable
from .custom import CustomMeasurable
from .accuracy import AccuracyMeasurable, MAEMeasurable, MAPEMeasurable
from .scalar_from_embedding import ScalarFromEmbeddingMeasurable
from .distance import DistanceMeasurable
from .reccomendation_hit_rate import RecHitRateMeasurable

from .measurable_resolver import MeasurableResolver



import numpy as np
from uptrain.core.classes.helpers import AnnotationHelper
from uptrain.core.lib.helper_funcs import read_json, write_json, cluster_and_plot_data
from uptrain.constants import AnnotationMethod


class DatasetHandler:
    """
    Handler class to help with all the dataset related functionalities

    ...
    Attributes
    ----------
    transformation_func : func
        Used to transform logged data-points to training compatible data-points
    annotation_method: AnnotationMethod
        Attached annotation method

    Methods
    -------
    add_annotations(dataset):
        Calls the Annotation Helper to add gt to retraining dataset
    set_transformation_func(func):
        Attach data transformation func to convert logged data-point -> Training dataset
    set_annotation_method(method):
        Set the annotation method
    """

    def __init__(self, framework=None, cfg=None, transformation_func=(lambda x: x)):
        self.transformation_func = transformation_func
        self.cluster_plot_func = cfg.cluster_visualize_func
        self.log_handler = framework.log_handler 
        np.random.seed(10)

    def add_annotations(self, dataset):
        """Adds gt to the dataset file by calling appropriate funcs on the annotation helper based on attached annotation method"""

        if self.annotation_method == AnnotationMethod.MASTER_FILE:
            annotated_data = self.annotation_helper.annotations_from_master_file(
                dataset
            )
        else:
            raise Exception("Unsupported annotation method")
        return annotated_data

    def set_transformation_func(self, func):
        """Attach data transformation func to convert logged data-point -> Training dataset"""

        self.transformation_func = func

    def set_annotation_method(self, method, args={}):
        """Set data annotation method"""

        self.annotation_method = method
        self.annotation_helper = AnnotationHelper(args=args)

    def transform_collected_data(self, data):
        """Transforms logged data via self.transformation_func"""
        if self.transformation_func:
            transformed_data = [self.transformation_func(x) for x in data]
            return transformed_data

    def create_retraining_dataset(
        self, dataset_location, new_data, old_dataset, ratio=5
    ):
        """Creates retraining dataset by combining original training data with collected data-points
        - Transforms logged data-points via transformation_func
        - Add annotatons via attached annotation method
        - Combine old training data and newly collected data into retraining dataset file
        """

        new_data = self.transform_collected_data(new_data)
        write_json(dataset_location + "/cleaned_dataset.json", new_data)

        if self.cluster_plot_func is not None:
            cluster_and_plot_data(
                np.array([x["kps"] for x in new_data]),
                9,
                cluster_plot_func=self.cluster_plot_func,
                plot_save_name=self.log_handler.get_plot_save_name("collected_edge_cases_clusters.png", "edge_cases"),
            )

        new_data = self.add_annotations(dataset_location + "/cleaned_dataset.json")
        write_json(dataset_location + "/annotated_dataset.json", new_data)
        write_json(
            dataset_location + "/training_dataset.json",
            self.merge_training_datasets(
                old_dataset, dataset_location + "/annotated_dataset.json", ratio=ratio
            ),
        )
        print(
            "Creating retraining dataset:",
            dataset_location + "/training_dataset.json",
            " by merging ",
            old_dataset,
            " and collected edge cases.\n",
        )

    def merge_training_datasets(self, old_dataset, new_dataset, ratio=1):
        """Merges two datasets with weights = specified ratio.
        Supports only integer ratios
        """

        old_data = read_json(old_dataset)
        new_data = read_json(new_dataset)
        new_data_upsampled = new_data * int(round(ratio))
        old_data.extend(new_data_upsampled)
        return old_data



from copy import deepcopy
import json
import os
import shutil
import pandas as pd
from datetime import datetime
import random
import numpy as np
from sklearn.preprocessing import normalize

from queue import SimpleQueue
from threading import Thread

from uptrain.core.classes.helpers import DatasetHandler, ModelHandler, config_handler
from uptrain.core.classes.logging import LogHandler
from uptrain.core.classes.managers import CheckManager
from uptrain.core.lib.helper_funcs import (
    read_json,
    extract_data_points_from_batch,
    add_data_to_warehouse,
    get_df_indices_from_ids,
    get_feature_names_list,
    load_list_from_df,
)

class Framework:
    """
    Base Framework class which handles model observability and retraining

    ...

    Attributes
    ----------
    fold_name : str
        Path to the folder where selected data-points are logged

    Methods
    -------
    smartly_add_data(inputs, outputs, extra_args=None):
        Checks if the given data-point represents an edge case
        and needs to be added to the retraining dataset.
    retrain():
        Creates a new version of the model by retraining
        on collected data.
    """

    def __init__(self, cfg_dict={}):
        """Initialises the uptrain Framework.

        Parameters
        ----------
        cfg : dict
            Config to initialize uptrain framework
        """

        cfg = config_handler.Config(**cfg_dict)
        self.run_background_log_consumer = cfg.run_background_log_consumer

        if self.run_background_log_consumer:
            self.msg_queue = SimpleQueue()
            cfg_copy = deepcopy(cfg_dict)
            cfg_copy.update({"run_background_log_consumer": False})
            daemon = Thread(target=background_log_consumer_task, args=(self.msg_queue,cfg_copy,), daemon=True, name='Background Log Consumer')
            daemon.start()
        else:
            training_args = cfg.training_args
            evaluation_args = cfg.evaluation_args

            self.orig_training_file = training_args.orig_training_file
            self.fold_name = cfg.retraining_folder
            if os.path.exists(self.fold_name):
                print("Deleting the folder: ", self.fold_name)
                shutil.rmtree(self.fold_name)
            os.mkdir(self.fold_name)

            self.log_data = cfg.logging_args.log_data
            self.use_cache = cfg.use_cache
            self.cache = {}
            self.predicted_count = 0
            self.extra_args = {}
            self.checks = cfg.checks
            self.batch_size = None
            self.feat_name_list = cfg.feat_name_list
            self.retrain_after = cfg.retrain_after
            self.data_id_type = cfg.data_id
            self.version = 0
            self.if_retraining = cfg.retrain
            self.path_all_data = os.path.join(self.fold_name, "all_data.csv")
            self.log_handler = LogHandler(framework=self, cfg=cfg)

            self.dataset_handler = DatasetHandler(framework=self, cfg=cfg)
            self.model_handler = ModelHandler()
            self.check_manager = CheckManager(self, self.checks)
            self.reset_retraining()

            if training_args.data_transformation_func:
                self.set_data_transformation_func(training_args.data_transformation_func)
            if training_args.annotation_method:
                am = training_args.annotation_method
                self.set_annotation_method(am.method, args=am.args)
            if evaluation_args.golden_testing_dataset:
                self.set_golden_testing_dataset(evaluation_args.golden_testing_dataset)
            if training_args.training_func:
                self.set_training_func(training_args.training_func)
            if evaluation_args.inference_func:
                self.set_inference_func(evaluation_args.inference_func)

    def reset_retraining(self):
        self.version += 1
        self.selected_count = 0
        self.smart_data_ids = []
        retrain_folder = os.path.join(self.fold_name, str(self.version))
        if not os.path.exists(retrain_folder):
            os.mkdir(retrain_folder)

    def get_data_id(self, inputs):
        if self.data_id_type in inputs:
            ids = inputs[self.data_id_type]
        elif self.data_id_type in inputs:
            ids = inputs[self.data_id_type]
        elif self.data_id_type == "utc_timestamp":
            timestamp = str(datetime.utcnow().timestamp()).replace(".", "")
            rand_int = random.sample(
                range(10 * self.batch_size), self.batch_size
            ).sort()
            ids = [int(timestamp + str(x)) for x in rand_int]
        elif self.data_id_type == "id":
            ids = list(
                range(self.predicted_count, self.predicted_count + self.batch_size)
            )
        else:
            raise Exception("Invalid Data id type: %s" % self.data_id_type)
        return ids

    def smartly_add_data(self, data, extra_args={}):
        """
        Checks if the given data-point is interesting.
        If yes, logs them to smart_data warehouse, which
        is used to create retraining dataset.
        """

        old_selected_count = self.selected_count
        smart_data = {}

        is_interesting, reasons = self.is_data_interesting(
            data, data["output"], data["gt"], extra_args=extra_args
        )
        num_selected_datapoints = np.sum(np.array(is_interesting))
        self.selected_count += num_selected_datapoints
        self.smart_data_ids.extend(np.array(data["id"])[np.array(is_interesting)])

        """
        Log only the interesting test cases to data 
        warehouse. Logged under sub-folder 'smart_data'
        """
        path_smart_data = os.path.join(
            self.fold_name, str(self.version), "smart_data.csv"
        )

        if self.log_data and (num_selected_datapoints > 0):
            data.update({"reasons": reasons})
            smart_data = extract_data_points_from_batch(
                data, np.where(is_interesting == True)[0]
            )
            add_data_to_warehouse(deepcopy(smart_data), path_smart_data)

        edge_cases_txt = str(self.selected_count) + " edge cases identified out of " + str(self.predicted_count) + " total samples"
        if self.selected_count > 0:
            self.log_handler.add_alert(
                "Number of edge cases collected",
                edge_cases_txt,
                "edge_cases"
            )

        if (not (self.selected_count == old_selected_count)) and (
            not (int(self.selected_count / 50) == int(old_selected_count / 50))
        ):
            print(edge_cases_txt)

    def infer_batch_size(self, inputs):
        batch_sizes = []
        for k, item in inputs.items():
            item_batch_size = len(item)
            batch_sizes.append(item_batch_size)
        if np.var(np.array(batch_sizes)) > 0:
            # TODO: Raise warning on what is going wrong
            raise Exception("Batch size should be same for all input features")
        return batch_sizes[0]

    def check_and_add_data(self, inputs, outputs, gts=None, extra_args={}):
        # inputs = dict(config_handler.InputArgs(**inputs))
        if ("id" in inputs) and (inputs["id"] is None):
            del inputs["id"]
        self.batch_size = self.infer_batch_size(inputs)

        ids = self.get_data_id(inputs)
        inputs.update({"id": ids})

        if self.feat_name_list is None:
            self.feat_name_list = get_feature_names_list(inputs)

        outputs = list(outputs) if outputs is not None else [None] * self.batch_size
        gts = list(gts) if gts is not None else [None] * self.batch_size
        data = inputs
        data.update(
            {
                "output": list(outputs),
                "gt": gts,
            }
        )

        if self.log_data:
            # Log all the data-points into all_data warehouse
            add_data_to_warehouse(deepcopy(data), self.path_all_data)

        # Check for any monitors
        self.check(data, extra_args)
        self.predicted_count += self.batch_size

        # Smartly add data for retraining
        self.smartly_add_data(data, extra_args)
        self.extra_args = extra_args
        return ids

    def check(self, data, extra_args={}):
        extra_args.update({"id": data["id"]})
        self.check_manager.check(
            data,
            data["output"],
            gts=data["gt"],
            extra_args=extra_args,
        )

    def is_data_interesting(self, inputs, outputs, gts=None, extra_args={}):
        """A data-point is deemed interesting if the defined signal is positive"""
        return self.check_manager.is_data_interesting(
            inputs, outputs, gts=gts, extra_args=extra_args
        )

    def set_data_transformation_func(self, func):
        """Attach data transformation func to convert
        logged data-point -> Training dataset"""
        self.dataset_handler.set_transformation_func(func)

    def set_annotation_method(self, method, args={}):
        """Attach data annotation pipeline"""
        self.dataset_handler.set_annotation_method(method, args=args)

    def need_retraining(self):
        """Checks if enough data-points are collected and the
        framework needs to kickoff model retraining"""
        if self.if_retraining:
            if self.selected_count > self.retrain_after:
                return True
        return False

    def retrain(self):
        """Retrains the model. Executes following steps sequentially.
        - Creates Retraining dataset by collecting data from the 'smart_data' warehouse
        - Retrains the model and saves it under the new version
        - Kicks off model comparison report
        - Deploys the new model and increments model version by 1
        """

        dataset_location = os.path.join(self.fold_name, str(self.version))
        print("\nKicking off re-training")
        # print(
        #     str(self.selected_count),
        #     "data-points selected out of " + str(self.predicted_count),
        # )

        # Collect newly collected data
        df = pd.read_csv(self.path_all_data)
        smart_data_indices = get_df_indices_from_ids(df, self.smart_data_ids)
        df_smart_data = df.loc[smart_data_indices]
        for feat in self.feat_name_list:
            df_smart_data[feat] = df_smart_data[feat].apply(json.loads)
        retraining_data = df_smart_data.to_dict(orient="records")

        # Generate training dataset
        self.dataset_handler.create_retraining_dataset(
            dataset_location, retraining_data, self.orig_training_file
        )
        # Retrain the model
        self.model_handler.retrain(
            dataset_location + "/training_dataset.json", self.version
        )
        print("Model retraining done...\n")
        print("Generating comparison report...")

        # Generate Model report
        self.model_handler.compare_new_model(
            self.golden_testing_dataset,
            self.version,
            self.orig_training_file,
            self.selected_count,
        )

        # TODO: Deploy new model
        self.reset_retraining()

    def set_training_func(self, func):
        """Attach model training pipeline"""
        self.model_handler.set_training_func(func)

    def set_golden_testing_dataset(self, dataset):
        """Attach testing dataset for model comparison"""
        self.golden_testing_dataset = dataset

    def set_inference_func(self, func):
        """Attach inference pipeline for model comparison"""
        self.model_handler.set_inference_func(func)

    def attach_ground_truth(self, gt_data):
        if type(gt_data) is str:
            gt_data = read_json(gt_data)
        elif type(gt_data) is not dict:
            print(f"Ground truth datatype {type(gt_data)} not supported.")
        gt_data = dict(config_handler.GroundTruthArgs(**gt_data))

        self.batch_size = len(gt_data["gt"])

        df = pd.read_csv(self.path_all_data)
        gt_id_indices = get_df_indices_from_ids(df, gt_data["id"])
        df.loc[gt_id_indices, "gt"] = np.array(gt_data["gt"], dtype="object")
        df.to_csv(self.path_all_data, index=False)

        """
        TODO: Currently assumes that input is only one column. 
        In some cases, the input might have multiple data 
        structures (e.g., cascaded models). 
        """
        df_gt = df.loc[gt_id_indices]
        data = dict(zip(
                self.feat_name_list,
                [load_list_from_df(df_gt, x) for x in self.feat_name_list],
            ))
        data.update({
            "output": load_list_from_df(df_gt, "output"),
            "id": list(gt_data["id"]),
            "gt": list(gt_data["gt"]),
        })
        self.check(data, extra_args=self.extra_args)
        self.smartly_add_data(data, extra_args=self.extra_args)

    #TODO: @Vipul - Do we need this?
    def feat_slicing(self, relevant_feat_list, limit_list):
        """
        This function checks anomalies for a subset of data.
        feat_name: Feature on which data slicing is done.
        lower_limit: Lower limit for feature value
        upper_limit: Upper limit for feature value
        """
        df = pd.read_csv(self.path_all_data)
        input_arr = np.array(load_list_from_df(df, "data"))

        # Create normalized array
        input_arr = normalize(input_arr, axis=1, norm="l1")
        relevant_ids_all = set(range(len(input_arr)))
        for i, feat_name in enumerate(relevant_feat_list):
            feat_id = self.feat_name_list.index(feat_name)
            relevant_ids = (input_arr[:, feat_id] >= limit_list[i][0]) & (
                input_arr[:, feat_id] <= limit_list[i][1]
            )
            relevant_ids_all = relevant_ids_all.intersection(set(relevant_ids))

        df_feat = df.loc[relevant_ids]
        self.batch_size = len(df_feat)

        data = {
            "data": load_list_from_df(df_feat, "data"),
            "output": load_list_from_df(df_feat, "output"),
            "id": load_list_from_df(df_feat, "id"),
            "gt": load_list_from_df(df_feat, "gt"),
        }
        self.extra_args.update({"feat_name": feat_name})
        self.check_manager = CheckManager(self, self.checks)
        self.check(data, extra_args=self.extra_args)

    def log_measurable(self, ids, vals, col_name):
        for idx in range(len(ids)):
            self.cache[col_name].update(
                {ids[idx]: extract_data_points_from_batch(vals, idx)}
            )

    def convert_inputs_table_to_dict(self, inputs):
        cols = inputs.columns.tolist()
        ids = inputs.index.tolist()
        data = {}
        for col in cols:
            data.update({col: np.array(list(inputs[col]))})
        if 'id' not in cols:
            data.update({"id": np.array(ids)})
        return data
    
    def convert_dict_values_to_numpy_values(self, inputs: dict) -> dict:
        data = {}
        for key, value in inputs.items():
            data.update({key: np.array(value)})
        return data

    def log(self, inputs=None, outputs=None, gts=None, identifiers=None, extra=None):
        if self.run_background_log_consumer:
            data = {"inputs": inputs, "outputs": outputs, "gts": gts, "identifiers": identifiers, "extra": extra}
            try:
                self.msg_queue.put(data)
            except:
                print("Model monitoring failed!")
            # Return identifiers by fetching it from the background thread
            return
        else:
            # if (inputs is not None) and (outputs is None):
            #     raise Exception("Predictions should be present while logging inputs")
            if (inputs is None) and (outputs is not None):
                raise Exception("Inputs should be present while logging predictions")
            if (gts is not None) and (identifiers is None):
                raise Exception("Identifiers should be present while logging ground truths")

            if inputs is not None:
                if isinstance(inputs, pd.DataFrame):
                    inputs = self.convert_inputs_table_to_dict(inputs)
                elif isinstance(inputs, dict):
                    inputs = self.convert_dict_values_to_numpy_values(inputs)
                else:
                    raise Exception("Inputs was expected to be a Pandas Dataframe or Python Dictionary")
                identifiers = self.check_and_add_data(inputs, outputs)

            if gts is not None:
                self.attach_ground_truth({"id": identifiers, "gt": gts})

            if self.need_retraining():
                self.retrain()

            return identifiers

    def clear_cache(self):
        self.cache = {}

def background_log_consumer_task(msg_queue, cfg):
    framework = Framework(cfg_dict=cfg)
    # run forever
    while True:
        # get the next message
        data = msg_queue.get()

        # log the message
        framework.log(inputs=data['inputs'], outputs=data['outputs'], gts=data['gts'], identifiers=data['identifiers'], extra=data['extra'])


import streamlit as st
import pandas as pd
from glob import glob
import os
import sys
import plotly.graph_objects as go
import numpy as np
import json
import plotly.express as px
import random
import streamlit.components.v1 as components
import shap


st.set_page_config(
    page_title="UpTrain Dashboard",
    layout="wide",
    page_icon='https://github.com/uptrain-ai/uptrain/raw/dashboard/uptrain/core/classes/logging/uptrain_logo_icon.png'
)
st.title("UpTrain Live Dashboard")
st_style = """<style> footer {visibility: hidden;} </style>"""
st.markdown(st_style, unsafe_allow_html=True)

# Getting the streamlit log folder
log_folder = sys.argv[1]

# Get metadata
metadata_file = os.path.join(log_folder, "metadata.json")
with open(metadata_file, encoding='utf-8') as f:
    metadata = json.loads(f.read())
model_args = metadata['model_args']
feature_args = metadata['feature_args']

model_to_compare = None
other_models = {}
num_models_compare = 1
if model_args:
    if len(model_args) > 1:
        # st.sidebar.subheader("Select model type to compare against")
        all_model_types = []
        for model_type in model_args:
            all_model_types.append(model_type['feature_name'])
        model_selected = st.sidebar.selectbox("Model type to compare with", 
            all_model_types, key='select_model_args')
        model_compare_ind = all_model_types.index(model_selected)
        
        st.sidebar.subheader("Select other model types")
        for i, model in enumerate(model_args):
            if i==model_compare_ind:
                continue
            model_name = model['feature_name']
            value = st.sidebar.selectbox(model_name, model['allowed_values'], 
                key=f'model_{model_name}')
            other_models.update({model_name: value})

    elif len(model_args) == 1:
        model_compare_ind = 0   
    model_to_compare = model_args[model_compare_ind]
    num_models_compare = len(model_to_compare['allowed_values'])

features_to_slice = {}
if feature_args:
    st.sidebar.subheader("Select relevant features")
    for i, feature in enumerate(feature_args):
        feature_name = feature['feature_name']
        allowed_feats = feature['allowed_values']
        allowed_feats.insert(0, 'All')
        value = st.sidebar.selectbox(feature_name, allowed_feats, 
            key=f'feature_{feature_name}')
        if value != 'All':
            features_to_slice.update({feature_name: value})


def return_plotly_fig(y_axis, x_axis="Num predictions", x_log=False, y_log=False):
    fig = go.Figure()
    fig.update_xaxes(title_text=x_axis)
    fig.update_yaxes(title_text=y_axis)
    if x_log:
        fig.update_xaxes(type="log")
    if y_log:
        fig.update_yaxes(type="log")
    return fig


def slice_data(
    df, 
    features_to_slice=None, 
    model_to_compare=None, 
    other_models={},
    j=0,
    ):
    cond = None
    if features_to_slice is not None:
        for feat_name, value in features_to_slice.items():
            if value != 'All':
                if 'feature_' + feat_name in df.columns:
                    if cond is None:
                        cond = (df['feature_' + feat_name] == value)
                    else:
                        cond = cond & (df['feature_' + feat_name] == value)
                else:
                    cond = [False] * len(df)
    if model_to_compare is not None:
        model = model_to_compare['allowed_values'][j]
        model_type = model_to_compare['feature_name']
        if 'model_' + model_type in df.columns:
            if cond is None:
                cond = (df['model_' + model_type] == model)
            else:
                cond = cond & (df['model_' + model_type] == model)
        else:
            cond = [False] * len(df)
    for model_name, value in other_models.items():
        if 'model_' + model_name in df.columns:
            if cond is None:
                cond = (df['model_' + model_name] == value)
            else:
                cond = cond & (df['model_' + model_name] == value)
        else:
            cond = [False] * len(df)
    if cond is not None:
        df = df[cond]
    return df


def plot_line_charts(files, plot_name):
    # Getting plot metadata from the first file
    if len(files) > 1000:
        files = random.choices(files, k=1000)
    df = pd.read_csv(files[0])

    for key in df.keys():
        if key.startswith("x_"):
            x_axis = key
        if key.startswith("y_"):
            y_axis = key

    col1, col2 = st.columns(2)
    with col1:
        x_log = st.checkbox(
            "log x", help="x-axis in log-scale", key=plot_name + "x"
        )
    with col2:
        y_log = st.checkbox(
            "log y", help="y-axis in log-scale", key=plot_name + "y"
        )
    
    cols = st.columns(2)
    for j in range(num_models_compare):
        fig = return_plotly_fig(y_axis, x_axis, x_log, y_log)
        for i, csv_file in enumerate(files):
            # Reading the csv file
            df = pd.read_csv(csv_file)
            
            df = slice_data(df, features_to_slice, model_to_compare, other_models, j)

            # Getting plot_id
            plot_id = os.path.split(csv_file)[-1].split(".")[0]
            fig = fig.add_trace(
                go.Scatter(
                    x=df[x_axis],
                    y=df[y_axis],
                    name=str(i) + "," + plot_id,
                )
            )

        with cols[j % 2]:
            if model_to_compare is not None:
                model_name = model_to_compare['allowed_values'][j]
                st.subheader(f'Model: {model_name}')
            st.plotly_chart(fig, use_container_width=True)


def plot_histograms(files, plot_name): 
    cols = st.columns(2)
    for j in range(num_models_compare):
        fig = go.Figure()
        for i, csv_file in enumerate(files):
            # Reading the csv file
            df = pd.read_csv(csv_file)
            df = slice_data(df, features_to_slice, model_to_compare, other_models, j)

            # Getting plot_id
            plot_id = os.path.split(csv_file)[-1].split(".")[0]
            df_y = df['y_points']
            if len(df_y) > 1000:
                df_y = np.random.choice(df_y, 1000)
            fig = fig.add_trace(go.Histogram(x=df_y, name=plot_id))

        with cols[j % 2]:
            if model_to_compare is not None:
                model_name = model_to_compare['allowed_values'][j]
                st.subheader(f'Model: {model_name}')
            st.plotly_chart(fig, use_container_width=True)


def plot_histogram(file):
    with open(file, encoding='utf-8') as f:
        data = json.loads(f.read())
    if isinstance(data, dict):
        fig = go.Figure()
        for key in data.keys():
            fig = fig.add_trace(go.Histogram(x=data[key], name=key))
    else:
        fig = go.Figure(data=[go.Histogram(x=data)])
    st.plotly_chart(fig)


def plot_umap(file, j=0):
    with open(file, encoding='utf-8') as f:
        data = json.loads(f.read())
    arr = np.array(data["umap"])
    clusters = data["clusters"]
    x = arr[:, 0]
    y = arr[:, 1]
    dictn = {'x': x, 'y': y, 'color': clusters}
    hover_data = []
    if "hover_texts" in data:
        dictn.update({'hover': data['hover_texts']})
        if len(data['hover_texts']):
            hover_data = list(data['hover_texts'][0].keys())

    if arr.shape[1] == 3:
        dictn.update({'z': arr[:, 2]})

    for key in data.keys():
        if key not in ['umap', 'clusters', 'hover_texts']:
            dictn.update({key: data[key]})
    df = pd.DataFrame(dictn)
    df = slice_data(df, features_to_slice, model_to_compare, other_models, j)

    if "hover" in list(df.columns):
        hover_df = pd.DataFrame(list(df['hover']))
        for key in hover_data:
            if key not in hover_df.columns:
                df[key] = [''] * len(hover_df)
            else:
                df[key] = pd.Series(hover_df[key]).fillna('').tolist()
    else:
        hover_df = pd.DataFrame()
    
    if arr.shape[1] == 2:
        fig = px.scatter(df, x='x', y='y', color='color', hover_data=hover_data)
    elif arr.shape[1] == 3:
        z = list(df['z'])
        fig = px.scatter_3d(df, x='x', y='y', z='z', color='color', hover_data=hover_data)
    else:
        raise ("Umap dimension not 2D or 3D.")
    st.plotly_chart(fig, use_container_width=True)
    # st.write(
    #     f"Number of clusters: {len(set(clusters))}"
    # )


def get_view_arr_from_files(files):
    view_arr = []
    for file in files:
        view_arr.append(int(os.path.split(file)[-1].split('_')[0]))
    view_arr.sort()
    return np.unique(view_arr)


def plot_umaps(files, plot_name, sub_dir):
    view_arr = get_view_arr_from_files(files)
    if len(view_arr > 0):
        selected_count = st.selectbox(f"Cluster View Point", view_arr, key=plot_name+'count')
    cols = st.columns(2)
    for j in range(num_models_compare):
        if model_to_compare is not None:
            model_compare_name = model_to_compare['allowed_values'][j]
            model_others_name = list(other_models.values())[0]
            file_name = str(selected_count) + '_' + model_compare_name + '_' + model_others_name + '.json'
            file_name = sub_dir + '/' + file_name
            with cols[j%2]:
                st.subheader(f'Model: {model_compare_name}, Signal: {model_others_name}, Count: {selected_count}')
                if os.path.exists(file_name):
                    plot_umap(file_name, j) 
                else:
                    st.write("Not sufficient data.")
        else:
            for file in files:
                count = os.path.split(file)[-1].split(".")[0]
                if int(count) < 0:
                    plot_umap(file, j) 
                else:
                    if st.checkbox(f"For count {count}", key=plot_name+str(count)):
                        plot_umap(file, j)  

def plot_bar(file):
    with open(file, encoding='utf-8') as f:
        data = json.loads(f.read())
    fig = go.Figure()
    for bar_name in data:
        if bar_name == "hover_text":
            continue
        bar_dict = data[bar_name]
        keys, values = zip(*bar_dict.items())
        if "hover_text" in data:
            hover_text = data['hover_text'][bar_name]
        else:
            hover_text = {}
        hover_text = list(hover_text.values())
        fig = fig.add_trace(
            go.Bar(
                x=list(keys),
                y=list(values),
                name=bar_name,
                hovertext=hover_text
            )
        )
    st.plotly_chart(fig)


def plot_for_count(files, plot_func, plot_name):
    for file in files:
        count = os.path.split(file)[-1].split(".")[0]
        if int(count) < 0:
            plot_func(file) 
        else:
            if st.checkbox(f"For count {count}", key=plot_name+str(count)):
                plot_func(file)          


def plot_dashboard(dashboard_name):
    st.header(f"Dashboard {dashboard_name}")
    sub_dirs = [path[0] for path in os.walk(os.path.join(log_folder, dashboard_name))]
    for sub_dir in sub_dirs:
        sub_dir_split = os.path.normpath(sub_dir).split(os.path.sep)
        c1 = sub_dir_split[-1] == "umap_and_clusters"
        c2 = sub_dir_split[-2] == "bar_graphs"
        c3 = sub_dir_split[-1] == "alerts"
        c4 = sub_dir_split[-1] == "tsne_and_clusters"
        if c1 or c2 or c3 or c4:
            ext = "*.json"
        else:
            ext = "*.csv"

        # Getting the list of relevant files in streamlit logs
        files = [file for path, _, _ in os.walk(sub_dir)
                    for file in glob(os.path.join(path, ext))]
        plot_name = sub_dir_split[-1]

        ######### Showing Alerts ###########

        if sub_dir_split[-1] == "alerts":  
            for file in files:
                alert_name = os.path.split(file)[-1].split(".")[0]
                f = open(file)
                alert = json.load(f)
                st.subheader(alert_name)
                st.markdown("##### " + alert)
                st.markdown("""---""")

        ######### Line Plots ###########

        elif sub_dir_split[-2] == "line_plots":
            if st.sidebar.checkbox(f"Line-plot for {plot_name}"):
                st.markdown(f"### Line chart for {plot_name}")
                plot_line_charts(files, plot_name)
                st.markdown("""---""")    

        # ######### Plotting histograms ###########

        elif sub_dir_split[-2] == "histograms":
            if plot_name == "umap_and_clusters":
                if st.sidebar.checkbox(f"UMAP for {plot_name}"):
                    st.markdown(f"### UMAP for {plot_name}")
                    if model_args is not None:
                        plot_umaps(files, plot_name, sub_dir)
                    else:
                        for file in files:
                            plot_umap(file)
                    st.markdown("""---""") 
            elif plot_name == "tsne_and_clusters":  
                if st.sidebar.checkbox(f"t-SNE for {plot_name}"):
                    st.markdown(f"### t-SNE for {plot_name}")
                    if model_args is not None:
                        plot_umaps(files, plot_name, sub_dir)
                    else:
                        for file in files:
                            plot_umap(file)
                    st.markdown("""---""")   
            else:
                if st.sidebar.checkbox(f"Histogram for {plot_name}"):
                    st.markdown(f"### Histogram for {plot_name}")
                    # plot_for_count(files, plot_histogram, plot_name) 
                    plot_histograms(files, plot_name)
                    st.markdown("""---""") 

        ######### Plotting Bar Graphs ###########

        elif sub_dir_split[-2] == "bar_graphs":
            if st.sidebar.checkbox(f"Bar graph for {plot_name}"):
                st.markdown(f"### Bar graph for {plot_name}")
                plot_for_count(files, plot_bar, plot_name) 
                st.markdown("""---""")  


        ######### Plotting Images ###########

        # elif sub_dir_split[-1] == "images":
        if True:
            png_files = [
                        file
                        for path, _, _ in os.walk(sub_dir)
                        for file in glob(os.path.join(path, "*.png"))
                    ]
            for i, png_file in enumerate(png_files):
                # Getting image name
                image_name = png_file.split("/")[-1].split(".")[0]
                st.subheader(image_name)
                st.image(png_file)


def st_shap(plot, height=None):
    shap_html = f"<head>{shap.getjs()}</head><body>{plot.html()}</body>"
    components.html(shap_html, height=height) 


@st.cache
def get_data_shap(path_all_data, num_points):
    import pickle
    file = open(metadata["path_shap_file"], 'rb')
    explainer = pickle.load(file)
    file.close()
    df = pd.read_csv(path_all_data)
    if len(df) >= num_points:
        df = df[0:num_points]
    else:
        st.text("Not sufficient data points for SHAP")
        return []
    data_ids = [eval(x) for x in df["id"]]
    df = df.drop(columns=['id', 'output', 'gt'])
    return explainer(df), data_ids


st.sidebar.title("Select dashboards to view")
dashboard_names = next(os.walk(log_folder))[1]
for dashboard_name in dashboard_names:
    if st.sidebar.checkbox(f"Dashboard: {dashboard_name}"):
        plot_dashboard(dashboard_name)
    st.sidebar.markdown("""---""")

if metadata.get("path_shap_file", None):
    if st.sidebar.checkbox(f"SHAP explainability"):
        st.header(f"SHAP Explanability")
        
        path_all_data = metadata["path_all_data"]

        num_points = metadata["shap_num_points"]
        
        shap_values, data_ids = get_data_shap(path_all_data, num_points)

        shap.initjs() # for visualization
        st.set_option('deprecation.showPyplotGlobalUse', False)

        st.subheader("Feature-wise importance")
        st.text("Feature \"dist\" has the biggest impact on ride time predictions.")
        cols = st.columns(2)
        with cols[0]:
            st.pyplot(shap.plots.bar(shap_values))

        st.markdown("""---""")

        st.subheader("Explainability for each data-point")
        cols = st.columns(2)
        with cols[0]:
            data_point = st.selectbox("Select data-point for explainability", data_ids)

        index = data_ids.index(data_point)
        shap_val = shap_values[index]
        pred = sum(shap_val.values) + shap_val.base_values
        st.text(f"The predicted value is {pred:.1f} compared to the mean value of {shap_val.base_values:.1f}.")
            
        cols = st.columns(2)
        with cols[0]:
            shap.plots.waterfall(shap_val)
            st.pyplot()






import numpy as np


class CosineDistance:
    def compute_distance(self, base, ref) -> float:
        base = np.array(base)
        ref = np.array(ref)
        if len(base.shape) > 1:
            base_norm = np.linalg.norm(base, axis=1)
        else:
            base_norm = np.abs(base)
        if len(ref.shape) > 1:
            ref_norm = np.linalg.norm(ref, axis=1)
        else:
            ref_norm = np.abs(ref)
        return np.array(
            [
                1 - np.dot(base[i], ref[i]) / (base_norm[i] * ref_norm[i])
                for i in range(base.shape[0])
            ]
        )



import numpy as np

from uptrain.core.classes.monitors import AbstractMonitor
from uptrain.core.classes.signals import SignalManager
from uptrain.constants import Monitor


class EdgeCase(AbstractMonitor):
    dashboard_name = "edge_cases"
    monitor_type = Monitor.EDGE_CASE

    def base_init(self, fw, check):
        self.signal_manager = SignalManager()
        self.signal_manager.add_signal_formulae(check['signal_formulae'])
        self.num_preds = 0
        self.num_selected = 0

    def base_check(self, inputs, outputs, gts=None, extra_args={}):
        return

    def base_is_data_interesting(self, inputs, outputs, gts=None, extra_args={}):
        is_interesting = self.signal_manager.evaluate_signal(
            inputs, outputs, gts=gts, extra_args=extra_args
        )
        self.num_preds += len(is_interesting)
        self.num_selected += sum(is_interesting)
        self.log_handler.add_scalars(
            "num_edge_cases",
            {"y_num_edge_cases": self.num_selected},
            self.num_preds,
            self.dashboard_name,
        )
        reasons = []
        for is_in in is_interesting:
            if is_in:
                reasons.append("Edge_case_collected_via_Signal")
            else:
                reasons.append("None")
        return is_interesting, reasons

    def need_ground_truth(self):
        return False



from uptrain.core.classes import Framework
from uptrain.core.classes.signals import Signal
from uptrain.core.classes.monitors import CustomMonitor
from uptrain.constants import (
    ModelSignal,
    AnnotationMethod,
    Monitor,
    DataDriftAlgo,
    BiasAlgo,
    MeasurableType,
    Statistic,
    Visual,
)
from uptrain.core.encoders import UpTrainEncoder



import numpy as np

from uptrain.core.classes.measurables import Measurable


class AccuracyMeasurable(Measurable):
    def __init__(self, framework) -> None:
        super().__init__(framework)

    def _compute(self, inputs=None, outputs=None, gts=None, extra=None) -> any:
        gts = np.array(gts)
        outputs = np.reshape(np.array(outputs), gts.shape)
        return np.equal(outputs, gts)

    def col_name(self):
        return "Accuracy" 


class MAEMeasurable(Measurable):
    def __init__(self, framework) -> None:
        super().__init__(framework)

    def _compute(self, inputs=None, outputs=None, gts=None, extra=None) -> any:
        gts = np.array(gts)
        outputs = np.reshape(np.array(outputs), gts.shape)
        return np.squeeze(np.abs(outputs - gts))

    def col_name(self):
        return "MAE"


class MAPEMeasurable(Measurable):
    def __init__(self, framework) -> None:
        super().__init__(framework)

    def _compute(self, inputs=None, outputs=None, gts=None, extra=None) -> any:
        gts = np.array(gts)
        outputs = np.reshape(np.array(outputs), gts.shape)
        return 100*np.squeeze(np.abs(np.divide(outputs - gts, gts)))

    def col_name(self):
        return "MAPE"    



import numpy as np
import json_fix

from uptrain.constants import MODEL_SIGNAL_TO_FN_MAPPING


class Signal:
    """
    Signal Class to identify edge cases

    ...
    Attributes
    ----------
    name : str
        User-friendly name of the signal
    fn : func
        Signal evaluation func. If evaluated as true, data-point is considered as interesting.
    is_model_signal: bool
        Model signals are defined as part of the framework. Hence, users don't need to define evaluation funcs for them.
    extra_args: dict
        Any signal-specific args (ex: THRESHOLD)

    Methods
    -------
    evaluate(inputs, outputs, extra_args={}):
        Evaluates on given inputs, outputs and return a bool
    """

    def __init__(self, name, fn=None, is_model_signal=False, extra_args={}):
        self.name = name
        if is_model_signal and (fn is None):
            if not name in MODEL_SIGNAL_TO_FN_MAPPING:
                raise Exception(
                    "Evalution fn for signal: %s is not defined in constants" % name
                )
            fn = MODEL_SIGNAL_TO_FN_MAPPING[name]
        self.fn = fn
        self.extra_args = extra_args
        self.children = []

    def __json__(self):
        return {"feature_name": str(self)}

    def base_evaluate(self, inputs, outputs, gts=None, extra_args={}):
        extra_args.update(self.extra_args)
        if self.fn is None:
            raise Exception("Evalution fn not defined for signal: ", self.name)
        return self.fn(inputs, outputs, gts=gts, extra_args=extra_args)

    def evaluate(self, inputs, outputs, gts=None, extra_args={}):
        res = self.base_evaluate(inputs, outputs, gts=gts, extra_args=extra_args)
        return np.array(res)

    def __invert__(self):
        return InvertSignal(self)

    def __add__(self, other):
        return AddSignal(self, other)

    def __mul__(self, other):
        return MulSignal(self, other)

    def __and__(self, other):
        return AndSignal(self, other)

    def __or__(self, other):
        return OrSignal(self, other)

    def __xor__(self, other):
        return XorSignal(self, other)

    def __gt__(self, other):
        return GreaterThanSignal(self, other)

    def __lt__(self, other):
        return LessThanSignal(self, other)

    def __eq__(self, other):
        return EqualToSignal(self, other)

    def __ge__(self, other):
        return GreaterEqualToSignal(self, other)

    def __le__(self, other):
        return LessEqualToSignal(self, other)

    def __ne__(self, other):
        return NotEqualToSignal(self, other)

    def __str__(self):
        txt = str(self.name)
        if len(self.children):
            txt += "("
            txt += ",".join([str(x) for x in self.children])
            txt += ")"
        return txt


class InvertSignal(Signal):
    def __init__(self, base):
        super().__init__("Invert")
        self.base = base
        self.children = [base]

    def base_evaluate(self, inputs, outputs, gts=None, extra_args={}):
        base_res = self.base.evaluate(inputs, outputs, gts=gts, extra_args=extra_args)
        return np.logical_not(base_res)


class OperatorSignal(Signal):
    name = ""

    def __init__(self, base, other):
        super().__init__(self.name)
        self.base = base
        self.other = other
        self.children = [base, other]

    def base_evaluate(self, inputs, outputs, gts=None, extra_args={}):
        base_res = self.base.evaluate(inputs, outputs, gts=gts, extra_args=extra_args)
        if isinstance(self.other, Signal):
            other_res = self.other.evaluate(
                inputs, outputs, gts=gts, extra_args=extra_args
            )
        else:
            other_res = [self.other] * len(base_res)
        return self.operator(base_res, other_res)


class AddSignal(OperatorSignal):
    name = "Sum"

    def operator(self, base_res, other_res):
        return base_res + other_res


class MulSignal(OperatorSignal):
    name = "Product"

    def operator(self, base_res, other_res):
        return base_res * other_res


class AndSignal(OperatorSignal):
    name = "And"

    def operator(self, base_res, other_res):
        return np.logical_and(base_res, other_res)


class OrSignal(OperatorSignal):
    name = "Or"

    def operator(self, base_res, other_res):
        return np.logical_or(base_res, other_res)


class XorSignal(OperatorSignal):
    name = "Xor"

    def operator(self, base_res, other_res):
        return np.logical_xor(base_res, other_res)


class GreaterThanSignal(OperatorSignal):
    name = "Greater Than"

    def operator(self, base_res, other_res):
        return np.greater(base_res, other_res)


class LessThanSignal(OperatorSignal):
    name = "Less Than"

    def operator(self, base_res, other_res):
        return np.less(base_res, other_res)


class EqualToSignal(OperatorSignal):
    name = "Equal"

    def operator(self, base_res, other_res):
        return np.equal(base_res, other_res)


class GreaterEqualToSignal(OperatorSignal):
    name = "Greater Equal To"

    def operator(self, base_res, other_res):
        return np.greater_equal(base_res, other_res)


class LessEqualToSignal(OperatorSignal):
    name = "Less Equal To"

    def operator(self, base_res, other_res):
        return np.less_equal(base_res, other_res)


class NotEqualToSignal(OperatorSignal):
    name = "Not Equal"

    def operator(self, base_res, other_res):
        return np.not_equal(base_res, other_res)



import numpy as np
import copy

from uptrain.core.classes.monitors import AbstractMonitor
from uptrain.constants import Monitor
from uptrain.core.lib.helper_funcs import read_json
from uptrain.core.lib.algorithms import estimate_earth_moving_cost
from uptrain.core.classes.measurables import MeasurableResolver
from uptrain.core.classes.algorithms import Clustering

class DataDrift(AbstractMonitor):
    dashboard_name = "data_drift"
    monitor_type = Monitor.DATA_DRIFT
    is_embedding = None
    mode = None
    NUM_BUCKETS = 20
    INITIAL_SKIP = 2000

    def model_choices(self, check):
        self.is_embedding = check.get('is_embedding', None)
        if self.is_embedding == None:
            return [{"is_embedding": False}, {"is_embedding": True}]
        else:
            return [{"is_embedding": check['is_embedding']}]

    def base_init(self, fw, check):
        self.reference_dataset = check["reference_dataset"]
        self.cluster_plot_func = fw.dataset_handler.cluster_plot_func
        self.save_edge_cases = check.get("save_edge_cases", True)
        self.NUM_BUCKETS = check.get("num_buckets", self.NUM_BUCKETS)
        self.INITIAL_SKIP = check.get("initial_skip", self.INITIAL_SKIP)
        self.outlier_idxs = check.get("outlier_idxs", [])
        self.do_low_density_check = check.get("do_low_density_check", False)
        self.hover_measurable = MeasurableResolver(check.get("hover_label_args", None)).resolve(fw)
        self.count = 0
        self.bucket_labelling_info = {
            'latest_idxs_added_to_clusters': np.array([None] * self.NUM_BUCKETS),
            'idxs_closest_to_clusters': np.array([None] * self.NUM_BUCKETS),
            'closest_idxs_distance': np.array([None] * self.NUM_BUCKETS),
            'hover_vals_for_reference_clusters': np.array([' '] * self.NUM_BUCKETS),
            'hover_vals_for_production_clusters': np.array([' '] * self.NUM_BUCKETS)
        }
        self.prod_dist_counts_arr = []
        clustering_args = {
            "num_buckets": self.NUM_BUCKETS,
            'is_embedding': self.is_embedding,
            'plot_save_name': self.log_handler.get_plot_save_name("training_dataset_clusters.png", self.dashboard_name),
            'cluster_plot_func': self.cluster_plot_func
        }
        self.clustering_helper = Clustering(clustering_args)
        if self.is_embedding:
            self.plot_name = "Embeddings_" + self.measurable.col_name()
            self.emd_threshold = check.get("emd_threshold", 1)
        else:
            self.plot_name = "Scalar_" + self.measurable.col_name()
        self.bucket_reference_dataset()

    def need_ground_truth(self):
        return False

    def check(self, inputs, outputs, gts=None, extra_args={}):
        if len(self.children) > 0:
            feats = self.measurable.compute_and_log(
                inputs=inputs, outputs=outputs, gts=gts, extra=extra_args
            )
            if sum(list(feats.shape)[1:])/max(1,len(list(feats.shape)[1:])) <= 1:
                self.children = [self.children[0]]
            [x.check(inputs, outputs, gts=gts, extra_args=extra_args) for x in self.children]
        else:
            self.count += len(extra_args["id"])

            self.feats = self.measurable.compute_and_log(
                inputs=inputs, outputs=outputs, gts=gts, extra=extra_args
            )
            feats_shape = list(self.feats.shape)
            feats_shape.insert(1, 1)
            if len(feats_shape) == 2:
                feats_shape.insert(2, 1)
            self.feats = np.reshape(self.feats, tuple(feats_shape))

            if self.is_embedding:
                self.feats = self.feats / np.expand_dims(self.max_along_axis, 0)

            self.this_datapoint_cluster, self.prod_dist_counts = self.clustering_helper.infer_cluster_assignment(self.feats, self.prod_dist_counts)
            #TODO: Extend for feature-wise clusters
            if (self.hover_measurable is not None) and self.is_embedding:
                hover_measurable_vals = np.array(self.hover_measurable.compute_and_log(
                    inputs=inputs, outputs=outputs, gts=gts, extra=extra_args
                ))
                uniq_clusters, uniq_indexs = np.unique(self.this_datapoint_cluster, return_index=True)
                query_indexs = np.zeros(self.NUM_BUCKETS)
                self.bucket_labelling_info['latest_idxs_added_to_clusters'][uniq_clusters] = uniq_indexs
                query_indexs[uniq_clusters] = uniq_indexs
                query_indexs = np.array(query_indexs, dtype=int)
                self.bucket_labelling_info['hover_vals_for_production_clusters'] = hover_measurable_vals[query_indexs]

            self.prod_dist_counts_arr.append(self.prod_dist_counts.copy())

            self.prod_dist = (
                self.prod_dist_counts_arr[-1]
                - self.prod_dist_counts_arr[
                    int(max(self.count - self.INITIAL_SKIP, 0) / len(extra_args["id"]))
                ]
            ) / min(self.count, self.INITIAL_SKIP)

            if self.ref_dist.shape[0] == 1:
                if self.is_embedding:
                    bar_graph_buckets = ["cluster_" + str(idx) for idx in range(len(list(np.reshape(self.clusters,-1))))]
                else:
                    bar_graph_buckets = list(np.reshape(self.clusters,-1))

                self.log_handler.add_bar_graphs(
                    self.plot_name,
                    {
                        "reference": dict(zip(bar_graph_buckets, list(np.reshape(np.array(self.ref_dist[0]), -1)))),
                        "production": dict(zip(bar_graph_buckets, list(np.reshape(np.array(self.prod_dist[0]), -1)))),
                    },
                    self.dashboard_name,
                    hover_data = {
                        "reference": dict(zip(bar_graph_buckets, list(self.bucket_labelling_info['hover_vals_for_reference_clusters']))),
                        "production": dict(zip(bar_graph_buckets, list(self.bucket_labelling_info['hover_vals_for_production_clusters'])))
                    }
                )

            if self.count < self.INITIAL_SKIP:
                self.drift_detected = False
                return
            else:
                self.psis = np.zeros(self.ref_dist.shape[0])
                self.costs = np.zeros(self.ref_dist.shape[0])
                drift_detected = False
                for idx in range(self.ref_dist.shape[0]):
                    if self.is_embedding:
                        self.costs[idx] = estimate_earth_moving_cost(
                            self.prod_dist[idx], self.ref_dist[idx], self.clusters[idx]
                        )
                        drift_detected = drift_detected or (self.costs[idx] > self.emd_threshold)
                    else:
                        this_psi = sum(
                            [
                                (self.prod_dist[idx][jdx] - self.ref_dist[idx][jdx])
                                * np.log(
                                    max(self.prod_dist[idx][jdx], 0.0001)
                                    / self.ref_dist[idx][jdx]
                                )
                                for jdx in range(self.ref_dist.shape[1])
                            ]
                        )
                        self.psis[idx] = this_psi
                        drift_detected = drift_detected or (this_psi > 0.3)
                self.drift_detected = drift_detected
                if self.drift_detected:
                    alert = "Data Drift last detected at " + str(self.count) + " for Embeddings with Earth moving distance = " + str(float(self.costs[0]))
                    self.log_handler.add_alert(
                        "Data Drift Alert 🚨",
                        alert,
                        self.dashboard_name
                    )

            if self.is_embedding:
                dict_emc = dict(
                    zip(
                        ["y_" + self.measurable.col_name() for x in range(self.costs.shape[0])],
                        [float(x) for x in list(self.costs)],
                    )
                )
                # dict_emc.update({"threshold": self.emd_threshold})
                self.log_handler.add_scalars(
                    self.measurable.col_name() + " - earth_moving_costs",
                    dict_emc,
                    self.count,
                    self.dashboard_name,
                    file_name="embeddings"
                )
                self.log_handler.add_scalars(
                    self.measurable.col_name() + " - earth_moving_costs",
                    {list(dict_emc.keys())[0]: self.emd_threshold},
                    self.count,
                    self.dashboard_name,
                    file_name="threshold"
                )
            else:
                dict_psi = dict(
                    zip(
                        ["y_" + self.measurable.col_name() + "_" + str(x) for x in range(self.psis.shape[0])],
                        [float(x) for x in list(self.psis)],
                    )
                )
                dict_psi.update({"threshold": 0.3})
                self.log_handler.add_scalars(
                    self.measurable.col_name() + " - population_stability_index_scalar",
                    dict_psi,
                    self.count,
                    self.dashboard_name,
                )

    def base_is_data_interesting(self, inputs, outputs, gts=None, extra_args={}):
        is_interesting = np.array([False] * len(extra_args["id"]))
        reasons = ["None"] * len(extra_args["id"])

        if self.do_low_density_check and (len(self.low_density_regions) > 0):
            dists_from_low_density_regions = np.min(np.sum(np.abs(self.feats - np.expand_dims(self.low_density_regions, axis=0)),axis=2),axis=1)
            min_cluster_var = np.min(self.cluster_vars)
            is_close = dists_from_low_density_regions < min_cluster_var
            is_interesting = np.logical_or(is_close, is_interesting)
            for lkdx in range(len(is_close)):
                if is_close[lkdx]:
                    reasons[lkdx] = "Lies_to_Low_Density_Regions_In_Training_Distribution"

        if len(self.outliers):
            dists_from_outliers = np.min(np.sum(np.abs(self.feats - np.expand_dims(self.outliers, axis=0)),axis=2),axis=1)
            min_cluster_var = np.min(self.cluster_vars)
            is_close = dists_from_outliers < min_cluster_var
            is_close[3] = True
            is_interesting = np.logical_or(is_close, is_interesting)
            for lkdx in range(len(is_close)):
                if is_close[lkdx]:
                    reasons[lkdx] = "Close_to_User_annotated_Outliers"

        if self.save_edge_cases and self.drift_detected and not isinstance(self.feats[0,0,0], str):
            feats_shape = list(self.feats.shape)
            del feats_shape[1]
            self.feats = np.reshape(self.feats, tuple(feats_shape))
            for idx in range(self.ref_dist.shape[0]):
                if self.is_embedding:
                    dist_from_cluster = (
                        np.sum(
                            np.abs(
                                self.clusters[idx][self.this_datapoint_cluster]
                                - self.feats
                            ),
                            axis=tuple(range(1, len(feats_shape))),
                        )
                        / self.cluster_vars[idx][self.this_datapoint_cluster]
                    )
                else:
                    dist_from_cluster = np.abs(
                        np.reshape(
                            self.clusters[idx][self.this_datapoint_cluster[idx]], -1
                        )
                        - self.feats[:, idx]
                    ) / np.reshape(
                        self.cluster_vars[idx][self.this_datapoint_cluster[idx]], -1
                    )
                for jdx in range(dist_from_cluster.shape[0]):
                    if dist_from_cluster[jdx] > 2:
                        if self.is_embedding:
                            min_dist_from_clusters = np.min(
                                np.sum(
                                    np.abs(self.clusters[idx] - self.feats[jdx]),
                                    axis=tuple(range(1, len(feats_shape))),
                                )
                                / self.cluster_vars[idx]
                            )
                        else:
                            min_dist_from_clusters = np.min(
                                np.abs(
                                    np.reshape(self.clusters[idx], -1)
                                    - self.feats[jdx, idx]
                                )
                                / np.reshape(self.cluster_vars[idx], -1)
                            )
                        is_interesting[jdx] = is_interesting[jdx] or (
                            min_dist_from_clusters > 1
                        )
                        if min_dist_from_clusters > 1:
                            reasons[jdx] = "Away_from_all_training_clusters"

        return is_interesting, reasons


    def bucket_reference_dataset(self):
        self.ref_dist = []
        self.prod_dist = []
        self.ref_dist_counts = []
        self.prod_dist_counts = []
        data = read_json(self.reference_dataset)
        all_inputs = np.array(
            [self.measurable.extract_val_from_training_data(x) for x in data]
        )
        all_inputs_shape = list(all_inputs.shape)
        if len(all_inputs_shape) == 1:
            all_inputs_shape.insert(1, 1)
            all_inputs = np.reshape(all_inputs, all_inputs_shape)

        clustering_results = self.clustering_helper.cluster_data(all_inputs)

        self.buckets = np.array(clustering_results['buckets'])
        self.clusters = np.array(clustering_results['clusters'])
        self.cluster_vars = np.array(clustering_results['cluster_vars'])

        self.ref_dist = np.array(clustering_results['dist'])
        self.ref_dist_counts = np.array(clustering_results['dist_counts'])
        self.max_along_axis = clustering_results['max_along_axis']
        self.low_density_regions = clustering_results['low_density_regions']

        if self.hover_measurable is not None:
            all_hover_vals = np.array(
                [self.hover_measurable.extract_val_from_training_data(x) for x in data]
            )
            hover_label_idxs = [x[0] for x in list(clustering_results['idxs_closest_to_cluster_centroids'].values())]
            self.bucket_labelling_info['hover_vals_for_reference_clusters'] = all_hover_vals[hover_label_idxs]

        self.prod_dist = np.zeros(self.ref_dist.shape)
        self.prod_dist_counts = np.zeros(self.ref_dist_counts.shape)

        self.prod_dist_counts_arr.append(copy.deepcopy(self.prod_dist_counts))
        if len(self.outlier_idxs):
            self.outliers = all_inputs[np.array(self.outlier_idxs)]
        else:
            self.outliers = np.array([])


import os
import json
import copy
import numpy as np
import csv
import pandas as pd
from collections import OrderedDict
from sklearn.cluster import KMeans

from uptrain.core.encoders.uptrain_encoder import UpTrainEncoder


def cluster_and_plot_data(
    data, num_clusters, cluster_plot_func=None, plot_save_name="", normalisation=None
):
    kmeans = KMeans(n_clusters=num_clusters, random_state=1, n_init=10)
    kmeans.fit(data)
    all_clusters = kmeans.cluster_centers_
    all_labels = kmeans.labels_

    counts = np.zeros(num_clusters)
    uniq_lbs, uniq_cts = np.unique(all_labels, return_counts=True)
    for idx in range(uniq_lbs.shape[0]):
        counts[uniq_lbs[idx]] = uniq_cts[idx]

    cluster_vars = []
    cluster_dictn = {}
    for idx in range(len(all_clusters)):
        this_elems = data[np.where(all_labels == idx)[0]]
        cluster_dictn.update({idx: this_elems})
        cluster_vars.append(
            np.mean(np.sum(np.abs(this_elems - all_clusters[idx]), axis=1))
        )

    density_arr = []
    min_var = np.min(cluster_vars)
    for idx in range(len(data)):
        cluster_centroid_dists = np.sum(np.abs(all_clusters - data[idx]), axis=1)
        closest_cluster_labels = np.where(cluster_centroid_dists < min_var * 4)
        num_close_points = 0
        for this_label in closest_cluster_labels[0]:
            this_elems = cluster_dictn[this_label]
            all_elems_dists = np.sum(np.abs(this_elems - data[idx]), axis=1)
            closest_points_idxs = np.where(all_elems_dists < 1.5 * min_var)
            num_close_points += len(closest_points_idxs[0])
        density_arr.append(num_close_points-1)
    density_arr = np.array(density_arr)

    idxs_closest_to_cluster_centroids = {}
    for idx in range(len(all_clusters)):
        this_elems = cluster_dictn[idx]
        all_elems_dists = np.sum(np.abs(this_elems - all_clusters[idx]), axis=1)
        closest_idxs = np.argsort(all_elems_dists)[0:3]
        idxs_closest_to_cluster_centroids.update({idx: closest_idxs})

    dictn = []
    for idx in range(len(all_clusters)):
        dictn.append(
            {
                "cluster": all_clusters[idx],
                "count": counts[idx],
                "var": cluster_vars[idx],
                "idxs_closest": idxs_closest_to_cluster_centroids[idx]
            }
        )
    dictn.sort(key=lambda x: x["count"], reverse=True)

    all_clusters = np.array([x["cluster"] for x in dictn])
    counts = np.array([x["count"] for x in dictn])
    cluster_vars = np.array([x["var"] for x in dictn])
    idxs_closest_to_cluster_centroids = dict(zip(range(len(dictn)), [x['idxs_closest'] for x in dictn]))

    if normalisation is not None:
        all_clusters_renormalised = copy.deepcopy(all_clusters) * normalisation
    else:
        all_clusters_renormalised = all_clusters
    if cluster_plot_func is not None:
        cluster_plot_func(all_clusters_renormalised, counts, plot_save_name=plot_save_name)
    return all_clusters, counts, cluster_vars, density_arr, idxs_closest_to_cluster_centroids


def add_data_to_warehouse(data, path_csv, row_update=False):
    """Stores the dictionary "data" at location path_csv"""

    for k in list(data.keys()):
        if row_update and k == "id":
            continue
        data[k] = [json.dumps(x, cls=UpTrainEncoder) for x in data[k]]
    if not os.path.exists(path_csv):
        pd.DataFrame(data).to_csv(path_csv, index=False)
    else:
        if row_update:
            df = pd.read_csv(path_csv)
            for k in list(data.keys()):
                if row_update and k == "id":
                    continue
                if k not in list(df.columns):
                    df[k] = None
                df.loc[get_df_indices_from_ids(df, data["id"]), k] = np.array(
                    data[k], dtype="object"
                )
            pd.DataFrame(df).to_csv(path_csv, index=False)
        else:
            pd.DataFrame(data).to_csv(path_csv, index=False, mode="a", header=False)

def combine_data_points_for_batch(data):
    # what could be the generic logic???
    if isinstance(data, list):
        joined = None
        for idx in range(len(data)):
            elem = data[idx]
            if isinstance(elem, dict):
                if joined is None:
                    joined = {}
                for key in list(elem.keys()):
                    if key not in joined:
                        joined.update({key: []})
                    joined[key].append(elem[key])
                    if idx == len(data) - 1:
                        joined[key] = np.squeeze(np.array(joined[key]),axis=1)
            elif isinstance(elem, list):
                if len(elem) == 1:
                    if joined is None:
                        joined = []
                    joined.append(np.array(elem))
                    if idx == len(data) - 1:
                        joined = np.squeeze(np.array(joined),axis=1)
                else:
                    print(data)
                    import pdb; pdb.set_trace()
                    raise Exception("Not implemented, please contact developers")
            elif isinstance(elem, np.ndarray):
                if joined is None:
                    joined = []
                joined.append(elem)
                joined = np.squeeze(np.array(joined), axis=1)
        return joined
    else:
        print(data)
        raise Exception("Not implemented, please contact developers")

def extract_data_points_from_batch(data, idxs):
    if isinstance(data, dict):
        this_data = {}
        for key in list(data.keys()):
            this_data.update({key: extract_data_points_from_batch(data[key], idxs)})
        return this_data
    elif isinstance(data, np.ndarray):
        return np.array(data[np.array(idxs)])
    elif isinstance(data, list):
        if isinstance(idxs, int):
            return data[idxs]
        else:
            return [data[x] for x in list(idxs)]
    else:
        return data


def get_feature_names_list(inputs):
    feat_name_list = []
    for k in list(inputs.keys()):
        if not (k == "id"):
            feat_name_list.append(k)
    return feat_name_list


def add_data_to_batch(data, this_data):
    """Add data point this_data to data"""
    if isinstance(data, dict):
        assert isinstance(this_data, dict)
        for key, val in this_data.items():
            data.update({key: val})
        return data
    else:
        Exception("Invalid Data id type: %s" % type(data))


def get_df_indices_from_ids(df, ids):
    if not isinstance(df['id'][0], str):
        all_id_array = np.array(df['id'])
        if np.all(np.diff(all_id_array) >= 0):
            return np.searchsorted(all_id_array, ids)
        else:
            sorter = np.argsort(all_id_array)
            return sorter[np.searchsorted(all_id_array, ids, sorter=sorter)]
    else:
        id_str_list = [eval(x) for x in df['id']]
        sorter = np.argsort(id_str_list)
        return sorter[np.searchsorted(id_str_list, ids, sorter=sorter)]


def read_json(file_name):
    with open(file_name) as f:
        data = json.load(f)
    return data


def write_json(file_name, data):
    with open(file_name, "w") as f:
        json.dump(data, f)


def write_csv_row(file_name, data):
    with open(file_name, "a") as f_object:
        writer_object = csv.writer(f_object)
        writer_object.writerow(data)
        f_object.close()


def read_csv(file_name):
    return pd.read_csv(file_name)


def load_list_from_df(df, column):
    try:
        out = [json.loads(x) for x in list(df[column])]
    except:
        out_json = [json.dumps(x) for x in list(df[column])]
        out = [json.loads(x) for x in out_json]
    return out



import numpy as np

from uptrain.core.classes.monitors import AbstractMonitor
from uptrain.constants import MeasurableType
from uptrain.core.classes.measurables import MeasurableResolver
from uptrain.constants import Monitor

class Accuracy(AbstractMonitor):
    dashboard_name = "accuracy"
    monitor_type = Monitor.ACCURACY
    
    def base_init(self, fw, check):
        if check.get("measurable_args", None):
            self.measurable = MeasurableResolver(check["measurable_args"]).resolve(fw)
            self.plot_name = check["measurable_args"]['type']
        else:
            self.measurable = MeasurableResolver(
                {"type": MeasurableType.MAE}).resolve(fw)
        self.abs_err_arr = np.array([])
        self.avg_acc = 0
        self.log_handler = fw.log_handler

    def need_ground_truth(self):
        return True
    
    def base_check(self, inputs, outputs, gts=None, extra_args={}):
        batch_abs_error = self.measurable.compute_and_log(inputs, outputs, gts, extra_args)
        self.abs_err_arr = np.append(self.abs_err_arr, batch_abs_error)

        self.log_handler.add_scalars(
            self.plot_name,
            {"y_" + self.plot_name: np.mean(self.abs_err_arr)},
            len(self.abs_err_arr),
            self.dashboard_name,
        )


import umap
import copy
from sklearn.cluster import DBSCAN
import numpy as np
from uptrain.core.lib.helper_funcs import cluster_and_plot_data

from uptrain.core.classes.visuals import AbstractVisual
from uptrain.constants import Visual, Statistic
from uptrain.core.classes.measurables import MeasurableResolver
from uptrain.core.lib.helper_funcs import read_json

class Umap(AbstractVisual):
    visual_type = Visual.UMAP
    dashboard_name = "umap_and_clusters"

    def base_init(self, fw, check):
        self.framework = fw
        self.label_measurable = MeasurableResolver(check.get("label_args", None)).resolve(
            fw
        )
        self.hover_measurables = [
            MeasurableResolver(x).resolve(fw) for x in check.get("hover_args", [])
        ]
        self.hover_names = [x.col_name() for x in self.hover_measurables]

        self.count_checkpoints = check.get("count_checkpoints", ["all"])
        self.min_dist = check.get("min_dist", 0.01)
        self.n_neighbors = check.get("n_neighbors", 20)
        self.metric_umap = check.get("metric_umap", "euclidean")
        self.dim = check.get("dim", '2D')
        self.min_samples = check.get("min_samples", 5)
        self.eps = check.get("eps", 2.0)
        self.total_count = 0
        self.prev_calc_at = 0
        self.update_freq = check.get('update_freq', 10000)
        self.vals = []
        self.labels = []
        self.hover_texts = []
        self.do_clustering = check.get("do_clustering", False)
        self.feature_dictn = {}

        self.initial_dataset = check.get('initial_dataset', None)
        if self.initial_dataset is not None:
            data = read_json(self.initial_dataset)
            self.vals.extend(
                [self.measurable.extract_val_from_training_data(x) for x in data]
            )
            if self.label_measurable is not None:
                self.labels.extend(
                    [self.label_measurable.extract_val_from_training_data(x) for x in data]
                )
            if len(self.hover_measurables):
                raw_data = np.abs([x['bert_embs_downsampled'] for x in data])
                self.max_along_axis = np.max(raw_data, axis=0)
                norm_data = raw_data/self.max_along_axis

                _, _, _, points_density, _ = cluster_and_plot_data(norm_data,20)

                offset = 0
                for data_point in data:
                    this_hovers = [x.extract_val_from_training_data(data_point) for x in self.hover_measurables]
                    self.hover_texts.append(dict(zip(self.hover_names, this_hovers)))
                    self.hover_texts[offset].update({"Training Density": int(points_density[offset])})
                    self.hover_texts[offset].update({"idx": offset})
                    offset += 1
            if len(self.feature_measurables):
                all_features = [list(np.reshape([feature_measurable.extract_val_from_training_data(x) for x in data], -1)) for feature_measurable in self.feature_measurables]
                this_dict = dict(zip(['feature_' + x for x in self.feature_names], all_features))
                for key in this_dict.keys():
                    if key in self.feature_dictn:
                        self.feature_dictn[key].extend(this_dict[key])
                    else:
                        self.feature_dictn.update({key: list(this_dict[key])})

    def base_check(self, inputs, outputs, gts=None, extra_args={}):
        if self.measurable is not None:
            vals = self.measurable.compute_and_log(
                inputs, outputs, gts=gts, extra=extra_args
            )
            self.vals.extend(vals)
        if self.label_measurable is not None:
            labels = self.label_measurable.compute_and_log(
                inputs, outputs, gts=gts, extra=extra_args
            )
            self.labels.extend(labels)

        if len(self.feature_measurables):
            all_features = [x.compute_and_log(
                inputs, outputs, gts=gts, extra=extra_args
            ) for x in self.feature_measurables]
            this_dict = dict(zip(['feature_' + x for x in self.feature_names], all_features))
            for key in this_dict.keys():
                if key in self.feature_dictn:
                    self.feature_dictn[key].extend(this_dict[key])
                else:
                    self.feature_dictn.update({key: list(this_dict[key])})

        if len(self.hover_measurables):
            hover_values = [x.compute_and_log(
                inputs, outputs, gts=gts, extra=extra_args
            ) for x in self.hover_measurables]
            hover_table = np.array(hover_values).transpose()
            self.hover_texts.extend([dict(zip(self.hover_names, values)) for values in hover_table])

        self.total_count += len(extra_args['id'])
        if not ((self.total_count - self.prev_calc_at) >= self.update_freq):
            return

        self.prev_calc_at = self.total_count
        models = dict(zip(['model_' + x for x in self.model_names], 
            [self.allowed_model_values[jdx][0] for jdx in range(len(self.model_names))]))
        for count in self.count_checkpoints:
            emb_list, label_list, hover_texts = self.get_high_dim_data(count)
            emb_list = np.array(emb_list)
            label_list = np.array(label_list)

            if emb_list.shape[0] > 10:
                clusters = []
                umap_list, clusters = self.get_umap_and_labels(
                    emb_list,
                    self.dim,
                    self.n_neighbors,
                    self.min_dist,
                    self.metric_umap,
                    self.eps,
                    self.min_samples,
                    label_list=label_list
                )
                this_data = {"umap": umap_list, "clusters": clusters}
                if len(hover_texts) > 0:
                    this_data.update({"hover_texts": hover_texts})
                self.log_handler.add_histogram(
                    "umap_and_clusters",
                    this_data,
                    self.dashboard_name,
                    models = models,
                    features = self.feature_dictn,
                    file_name = str(count) + "_" + '_'.join(list(models.values()))
                )

    def get_high_dim_data(self, count):
        if self.measurable is None:
            distribution_anomaly = list(
                filter(
                    lambda x: x.statistic_type == Statistic.DISTRIBUTION_STATS,
                    self.framework.check_manager.statistics_to_check,
                )
            )[0]
            data_dict = distribution_anomaly.get_feats_for_clustering(count, self.allowed_model_values)
            chosen_label_key = None
            chosen_hover_key = None
            if len(data_dict):
                temp_val = list(data_dict.values())[0]
                if len(temp_val):
                    temp_keys = list(temp_val.keys())
                    temp_keys = list(filter(lambda x: "visual_label_" in x, temp_keys))
                    if len(temp_keys) > 1:
                        print("Have multiple labels - " + str(temp_keys) + " .Using " + temp_keys[0] + " for labeling UMAPs." )
                    if len(temp_keys) > 0:
                        chosen_label_key = temp_keys[0]
                if len(temp_val):
                    temp_keys = list(temp_val.keys())
                    temp_keys = list(filter(lambda x: "visual_hover_text_" in x, temp_keys))
                    if len(temp_keys) > 1:
                        print("Have multiple hover texts - " + str(temp_keys) + " .Using " + temp_keys[0] + " for hovering UMAPs." )
                    if len(temp_keys) > 0:
                        chosen_hover_key = temp_keys[0]
            vals = np.array([data_dict[x]['val'] for x in data_dict])
            if vals.shape[0] > 0:
                vals = np.squeeze(vals, axis=1)
            labels = []
            if chosen_label_key is not None:
                labels = [data_dict[x][chosen_label_key] for x in data_dict]

            hover_texts = []
            if chosen_hover_key is not None:
                hover_texts = [{chosen_hover_key: data_dict[x][chosen_hover_key]} for x in data_dict]
            return vals, labels, hover_texts
        else:
            return self.vals, self.labels, self.hover_texts

    def get_umap_and_labels(
        self,
        emb_list,
        dim,
        n_neighbors,
        min_dist,
        metric,
        eps,
        min_samples,
        label_list=None
    ):
        if dim == "2D":
            n_components = 2
        else:
            n_components = 3


        umap_embeddings = umap.UMAP(
            n_neighbors=n_neighbors,
            n_components=n_components,
            min_dist=min_dist,
            metric=metric,
        ).fit_transform(emb_list)

        # Do DBSCAN clustering
        if self.do_clustering:
            clustering = DBSCAN(eps=eps, min_samples=min_samples).fit(umap_embeddings)
            labels = clustering.labels_
        else:
            labels = label_list
        labels = np.squeeze(np.array(labels))
        return umap_embeddings, labels



from enum import Enum
from uptrain.core.lib.model_signal_funcs import *


class ModelSignal(Enum):
    DEFAULT = 1
    CROSS_ENTROPY_CONFIDENCE = 2
    BINARY_ENTROPY_CONFIDENCE = 3
    PASS_ALL = 4


MODEL_SIGNAL_TO_FN_MAPPING = {
    ModelSignal.DEFAULT: pass_none,
    ModelSignal.CROSS_ENTROPY_CONFIDENCE: cross_entropy_confidence,
    ModelSignal.BINARY_ENTROPY_CONFIDENCE: binary_entropy_confidence,
    ModelSignal.PASS_ALL: pass_all,
}


class AnnotationMethod(Enum):
    MASTER_FILE = 1


class Monitor(str, Enum):
    ACCURACY = "accuracy"
    EDGE_CASE = "edge_case"
    DATA_DRIFT = "data_drift"
    CUSTOM_MONITOR = "custom_monitor"
    CONCEPT_DRIFT = "concept_drift"
    POPULARITY_BIAS = "popularity_bias"
    DATA_INTEGRITY = "data_integrity"


class Statistic(str, Enum):
    DISTANCE = "distance"
    DISTRIBUTION_STATS = "distribution_stats"
    CONVERGENCE_STATS = "convergence_stats"


class Visual(str, Enum):
    UMAP = "UMAP"
    TSNE = "t-SNE"
    SHAP = "SHAP"


class DataDriftAlgo(str, Enum):
    DDM = "DDM"
    ADWIN = "ADWIN"


class BiasAlgo(str, Enum):
    POPULARITY_BIAS = "popularity_bias"


class MeasurableType(str, Enum):
    INPUT_FEATURE = "input_feature"
    PREDICTION = "prediction"
    CUSTOM = "custom"
    ACCURACY = "accuracy"
    MAE = "MAE"
    MAPE = "MAPE"
    CONDITION_ON_INPUT = "condition_on_input"
    CONDITION_ON_PREDICTION = "condition_on_prediction"
    SCALAR_FROM_EMBEDDING = "scalar_from_embedding"
    DISTANCE = "distance"
    REC_HIT_RATE = "hit_rate"



from uptrain.core.classes.measurables import Measurable


class OutputFeatureMeasurable(Measurable):
    def __init__(self, framework, feature_name) -> None:
        super().__init__(framework)
        self.feature_name = feature_name

    def _compute(self, inputs=None, outputs=None, gts=None, extra=None) -> any:
        return outputs

    def col_name(self):
        return str(self.feature_name)

    # TODO: Decommission and find a generic way
    def extract_val_from_training_data(self, x):
        return x[self.feature_name]



from uptrain.core.classes.measurables import (
    Measurable,
    InputFeatureMeasurable,
)


class ScalarFromEmbeddingMeasurable(Measurable):
    def __init__(self, framework, idx, extract_from_args) -> None:
        super().__init__(framework)
        self.idx = idx
        self.extract_from = InputFeatureMeasurable(
            framework, extract_from_args["feature_name"]
        )

    def _compute(self, inputs=None, outputs=None, gts=None, extra=None) -> any:
        return self.extract_from._compute(
            inputs=inputs, outputs=outputs, gts=gts, extra=extra
        )[self.idx]

    def col_name(self):
        return (
            "Scalar [Index:"
            + str(self.idx)
            + "] from Embedding: "
            + str(self.extract_from.col_name())
        )

    # TODO: Decommission and find a generic way
    def extract_val_from_training_data(self, x):
        return self.extract_from.extract_val_from_training_data(x)[self.idx]



import numpy as np


def cross_entropy_confidence(inputs, outputs, gts=None, extra_args={}):
    logits = np.reshape(np.array(outputs), (len(outputs), -1))
    logits = logits - np.max(logits, axis=1)
    logits = np.exp(logits)
    logits = logits / np.sum(logits, axis=1)
    conf = np.max(logits, axis=1)
    return conf


def binary_entropy_confidence(inputs, outputs, gts=None, extra_args={}):
    conf = np.reshape(np.array(outputs), -1)
    conf = np.greater_equal(conf, 0.5) * conf + np.less(conf, 0.5) * (1 - conf)
    return conf


def pass_all(inputs, outputs, gts=None, extra_args={}):
    return True


def pass_none(inputs, outputs, gts=None, extra_args={}):
    return False



import numpy as np

from uptrain.core.classes.monitors import AbstractCheck

class AbstractMonitor(AbstractCheck):
    monitor_type = None

    def need_ground_truth(self):
        return False

    def is_data_interesting(self, inputs, outputs, gts=None, extra_args={}):
        if len(self.children) > 0:
            is_interesting = np.array([False] * len(extra_args["id"]))
            reasons = ["None"] * len(extra_args["id"])
            is_interesting = []
            reasons = []
            for child in self.children:
                    res = child.is_data_interesting(
                            inputs, outputs, gts=gts, extra_args=extra_args
                        )
                    is_interesting.append(res[0])
                    reasons.append(res[1])
            if len(reasons):
                final_reason = ["None"] * len(reasons[0])
                for reas in reasons:
                    for jdx in range(len(reas)):
                        if not (reas[jdx] == "None"):
                            final_reason[jdx] = reas[jdx]
            else:
                final_reason = []
            return np.greater(np.sum(np.array(is_interesting), axis=0), 0), np.array(final_reason)
        else:
            return self.base_is_data_interesting(inputs, outputs, gts=gts, extra_args=extra_args)

    def base_is_data_interesting(self, inputs, outputs, gts=None, extra_args={}):
        return np.array([False] * len(extra_args["id"])), np.array(['None'] * len(extra_args["id"]))



import numpy as np

from uptrain.core.classes.distances import DistanceResolver
from uptrain.core.classes.measurables import Measurable
from uptrain.core.classes.measurables import InputFeatureMeasurable


class DistanceMeasurable(Measurable):
    def __init__(self, framework, base, reference, distance_types) -> None:
        super().__init__(framework)
        self.base = InputFeatureMeasurable(framework, base["feature_name"])
        self.reference = reference
        self.ref_val = None
        self.distance_types = distance_types
        self.dist_classes = [DistanceResolver().resolve(x) for x in self.distance_types]

    def _compute(self, inputs=None, outputs=None, gts=None, extra=None) -> any:
        base_val = self.base.compute_and_log(
            inputs=inputs, outputs=outputs, gts=gts, extra=extra
        )
        if self.ref_val is None:
            self.ref_val = base_val

        dists = dict(
            zip(
                self.distance_types,
                [x.compute_distance(base_val, self.ref_val) for x in self.dist_classes],
            )
        )

        if self.reference == "running_diff":
            self.ref_val = base_val
        return dists

    def col_name(self, return_str=True):
        col_names = [
            "Distance "
            + str(self.base.col_name())
            + self.reference
            + str(x)
            for x in self.distance_types
        ]
        if return_str:
            col_names = str(col_names)
        return col_names



import json
import numpy as np
import datetime

class UpTrainEncoder(json.JSONEncoder):
    """Special json encoder for numpy types"""

    def default(self, obj):
        if isinstance(obj, datetime.datetime):
            return str(obj)
        if isinstance(obj, np.integer):
            return int(obj)
        elif isinstance(obj, np.floating):
            return float(obj)
        elif isinstance(obj, np.ndarray):
            return obj.tolist()
        return json.JSONEncoder.default(self, obj)



import numpy as np

from uptrain.core.lib.helper_funcs import (
    extract_data_points_from_batch,
    combine_data_points_for_batch,
)
from uptrain.core.classes.measurables import AbstractMeasurable


class Measurable(AbstractMeasurable):
    def __init__(self, framework) -> None:
        super().__init__()
        self.framework = framework
        self.cache_present_or_not = None

    def compute_and_log(self, inputs=None, outputs=None, gts=None, extra=None) -> None:
        if self.framework.use_cache:
            if self.cache_present_or_not is None:
                self.cache_present_or_not = 1
                self.framework.cache.update({self.col_name(): {}})

            compute_masks = []
            compute_idxs = []
            non_compute_vals = []
            for idx in range(len(extra["id"])):
                id = extra["id"][idx]
                if id in self.framework.cache[self.col_name()]:
                    compute_masks.append(0)
                    non_compute_vals.append(self.framework.cache[self.col_name()][id])
                else:
                    compute_masks.append(1)
                    compute_idxs.append(idx)

            compute_masks = np.array(compute_masks)
            compute_idxs = np.array(compute_idxs)

            compute_vals = self._compute(
                inputs=extract_data_points_from_batch(inputs, compute_idxs),
                outputs=extract_data_points_from_batch(outputs, compute_idxs),
                gts=extract_data_points_from_batch(gts, compute_idxs),
                extra=extract_data_points_from_batch(extra, compute_idxs),
            )

            # TODO: Replace below for loop by numpy operations
            vals = []
            compute_offset = 0
            non_compute_offset = 0
            for idx in range(len(extra["id"])):
                if compute_masks[idx] == 1:
                    vals.append(
                        extract_data_points_from_batch(compute_vals, [compute_offset])
                    )
                    compute_offset += 1
                else:
                    vals.append(
                        extract_data_points_from_batch(
                            non_compute_vals, [non_compute_offset]
                        )
                    )
                    non_compute_offset += 1
            vals = combine_data_points_for_batch(vals)
            try:
                self._log(extra["id"], vals)
            except:
                raise("Error in logging")
        else:
            vals = self._compute(inputs=inputs, outputs=outputs, gts=gts, extra=extra)
        return vals

    def _log(self, ids, vals) -> None:
        self.framework.log_measurable(ids, vals, self.col_name())



import numpy as np


def estimate_earth_moving_cost(prod_dist, ref_dist, clusters):
    num_clusters = clusters.shape[0]
    cost = 0
    for jdx in range(num_clusters):
        dirt_required = prod_dist[jdx] - ref_dist[jdx]
        dictn = []
        for kdx in range(num_clusters):
            if jdx == kdx:
                dirt_can_be_transported = 0
            else:
                dirt_can_be_transported = prod_dist[kdx] - ref_dist[kdx]
            this_dirt = -dirt_can_be_transported
            if this_dirt * dirt_required < 0:
                cost_per_dirt = 1000000000
            else:
                cost_per_dirt = np.sum(np.abs(clusters[kdx] - clusters[jdx]))
            dictn.append(
                {
                    "dirt": np.abs(this_dirt),
                    "idx": kdx,
                    "cost_per_dirt": cost_per_dirt,
                }
            )
        dictn = sorted(dictn, key=lambda x: x["cost_per_dirt"])
        this_cost = 0
        dirt_required = np.abs(dirt_required)
        for kdx in range(len(dictn)):
            if dirt_required > 0:
                this_cost += (
                    min(dictn[kdx]["dirt"], dirt_required)
                    * dictn[kdx]["cost_per_dirt"]
                )
                dirt_required -= min(dictn[kdx]["dirt"], dirt_required)
            else:
                break
        cost += this_cost
    return cost



import copy
from abc import ABC

from uptrain.core.classes.measurables import MeasurableResolver

class AbstractCheck(ABC):

    def __init__(self, fw, check) -> None:
        super().__init__()
        self.log_handler = fw.log_handler
        self.measurable = MeasurableResolver(check.get("measurable_args", None)).resolve(fw)

        self.feature_measurables = [
            MeasurableResolver(x).resolve(fw) for x in check.get("feature_args", [])
        ]
        self.feature_names = [x.col_name() for x in self.feature_measurables]

        self.model_measurables = [
            MeasurableResolver(x).resolve(fw) for x in check.get("model_args", [])
        ]
        self.model_names = [x.col_name() for x in self.model_measurables]

        all_choices = self.model_choices(check)
        self.children = []
        if len(all_choices) > 1:
            for choice in all_choices:
                check_copy = copy.deepcopy(check)
                check_copy.update(choice)
                self.children.append(self.__class__(fw, check_copy))
        else:
            self.base_init(fw, check)

    def base_init(self, fw, check):
        print(self.__class__)
        raise Exception("Should be defined for each class")

    def model_choices(self, check):
        allowed_model_values = [x['allowed_values'] for x in check.get('model_args', [])]
        num_model_options = sum([len(x) > 1 for x in allowed_model_values])

        all_choices = []
        if num_model_options > 0:
            for m in allowed_model_values[0]:
                check_copy = copy.deepcopy(check)
                check_copy['model_args'][0]['allowed_values'] = [m]
                check_copy['model_args'].append(copy.deepcopy(check_copy['model_args'][0]))
                del check_copy['model_args'][0]
                all_choices.extend(self.model_choices(check_copy))
            return all_choices
        else:
            self.allowed_model_values = [x['allowed_values'] for x in check.get('model_args', [])]
            return [{"model_args": check.get("model_args", [])}]

    def check(self, inputs, outputs, gts=None, extra_args={}):
        if len(self.children) > 0:
            [x.check(inputs, outputs, gts=gts, extra_args=extra_args) for x in self.children]
        else:
            self.base_check(inputs, outputs, gts=gts, extra_args=extra_args)

    def base_check(self, inputs, outputs, gts=None, extra_args={}):
        print(self.__class__)
        raise Exception("Should be defined for each class")


import os
import shutil
import urllib3
import json

class LogHandler:
    def __init__(self, framework=None, cfg=None):
        self.fw = framework
        log_folder = cfg.logging_args.log_folder
        self.path_all_data = framework.path_all_data

        if os.path.exists(log_folder):
            print("Deleting the folder: ", log_folder)
            shutil.rmtree(log_folder)

        self.st_writer = None
        if cfg.logging_args.st_logging:
            from uptrain.core.classes.logging.log_streamlit import StreamlitLogs

            self.st_log_folder = os.path.join(log_folder, "st_data")
            os.makedirs(self.st_log_folder, exist_ok=True)
            self.st_writer = StreamlitLogs(self.st_log_folder, port=cfg.logging_args.dashboard_port)

        # Get Webhook URL for alerting on slack
        self.webhook_url = cfg.logging_args.slack_webhook_url

        # Saving config to get model metadata
        self.cfg_metadata = {}
        if len(cfg.checks) > 0:
            check = cfg.checks[0]
            self.add_st_metadata({'model_args': check.get('model_args', None), 
                                  'feature_args': check.get('feature_args', None)})
        else:
            self.add_st_metadata({'model_args': None, 'feature_args': None})

    def add_st_metadata(self, new_dict):
        if self.st_writer is None:
            return
        self.cfg_metadata.update(new_dict)
        metadata_file = os.path.join(self.st_log_folder, "metadata.json")
        with open(metadata_file, "w") as f:
            json.dump(self.cfg_metadata, f)

    def get_plot_save_name(self, plot_name, dashboard_name):
        if self.st_writer:
            dir_name = os.path.join(self.st_log_folder, dashboard_name)
            os.makedirs(dir_name, exist_ok=True)
            return os.path.join(dir_name, plot_name)
        else:
            return ""

    def add_scalars(self, plot_name, dictn, count, dashboard_name, features={}, models={}, file_name=None, update_val=False):
        if self.st_writer is None:
            return
        dashboard_name, plot_name = self.dir_friendly_name(
            [dashboard_name, plot_name]
        )
        dictn.update(features)
        dictn.update(models)
        new_dictn = dict(
            zip(
                self.dir_friendly_name(list(dictn.keys())),
                dictn.values(),
            )
        )
        dashboard_dir = os.path.join(self.st_log_folder, dashboard_name)
        plot_folder = os.path.join(dashboard_dir, "line_plots", plot_name)
        os.makedirs(plot_folder, exist_ok=True)
        new_dictn.update({"x_count": count})
        if file_name is None:
            file_name = plot_name
        self.st_writer.add_scalars(new_dictn, plot_folder, file_name=file_name, update_val=update_val)

    def add_histogram(self, plot_name, data, dashboard_name, features=None, models=None, file_name=None):
        dashboard_name, plot_name = self.dir_friendly_name(
            [dashboard_name, plot_name]
        )
        if self.st_writer:
            dashboard_dir = os.path.join(self.st_log_folder, dashboard_name)
            plot_folder = os.path.join(dashboard_dir, "histograms", plot_name)
            os.makedirs(plot_folder, exist_ok=True)
            if file_name is None:
                file_name = plot_name
            self.st_writer.add_histogram(data, plot_folder, features=features, models=models, file_name=file_name)

    def add_bar_graphs(self, plot_name, data, dashboard_name, count=-1, hover_data={}):
        if self.st_writer is None:
            return
        dashboard_name, plot_name = self.dir_friendly_name(
            [dashboard_name, plot_name]
        )
        dashboard_dir = os.path.join(self.st_log_folder, dashboard_name)
        plot_folder = os.path.join(dashboard_dir, "bar_graphs", plot_name)
        os.makedirs(plot_folder, exist_ok=True)
        self.st_writer.add_bar_graphs(data, plot_folder, count, hover_data=hover_data)

    def dir_friendly_name(self, arr):
        if isinstance(arr, str):
            return self.dir_friendly_name([arr])[0]

        new_arr = [self.convert_str(x) for x in arr]
        return new_arr

    def convert_str(self, txt):
        txt = txt.replace("(", "_")
        txt = txt.replace(")", "_")
        txt = txt.replace(":", "_")
        txt = txt.replace(" ", "_")
        txt = txt.replace(",", "_")
        txt = txt.replace(">", "_")
        txt = txt.replace("<", "_")
        txt = txt.replace("=", "_")
        txt = txt.replace("-", "_")
        return txt

    def add_alert(self, alert_name, alert, dashboard_name):
        if self.st_writer:
            dashboard_name = self.dir_friendly_name(dashboard_name)
            dashboard_dir = os.path.join(self.st_log_folder, dashboard_name)
            plot_folder = os.path.join(dashboard_dir, "alerts")
            os.makedirs(plot_folder, exist_ok=True)
            self.st_writer.add_alert(alert_name, alert, plot_folder)

        if self.webhook_url:
            message = f"Dashboard: {dashboard_name}, Alert name: {alert_name}, Alert: {alert}"
            self.slack_notification({'text': message})

    def slack_notification(self, message):
        try:
            http = urllib3.PoolManager()
            response = http.request('POST',
                                    self.webhook_url,
                                    body = json.dumps(message),
                                    headers = {'Content-Type': 'application/json'},
                                    retries = False)
        except Exception as e:
            print("Caught Exception")
            print(e)



import numpy as np
import copy
import random

from uptrain.core.lib.helper_funcs import extract_data_points_from_batch
from uptrain.core.classes.distances import DistanceResolver
from uptrain.core.classes.statistics import AbstractStatistic
from uptrain.core.classes.measurables import MeasurableResolver
from uptrain.constants import Statistic


class Distribution(AbstractStatistic):
    dashboard_name = "distribution_stats"
    statistic_type = Statistic.DISTRIBUTION_STATS

    def base_init(self, fw, check):
        self.aggregate_measurable = MeasurableResolver(check["aggregate_args"]).resolve(
            fw
        )
        self.count_measurable = MeasurableResolver(check["count_args"]).resolve(
            fw
        )

        self.item_counts = {}

        self.distance_types = check["distance_types"]
        # ex: ['cosine_distance']

        self.count_checkpoints = np.unique(np.array([0] + check["count_checkpoints"]))
        # ex: array([    0,   200,   500,  1000,  5000, 20000])

        self.vals_dictn = dict(zip(list(self.count_checkpoints), [{} for x in list(self.count_checkpoints)]))
        # ex: {0: {agg_id_0: val_0, ..}, 200: {}, 500: {}, 1000: {}, 5000: {}, 20000: {}}

        self.distances_dictn = dict(zip(list(self.count_checkpoints), [dict(zip(self.distance_types, [[] for x in list(self.distance_types)])) for y in list(self.count_checkpoints)]))
        #ex: {0: {'cosine_distance': [d1, d2, ..]}, 200: {'cosine_distance': []}, 500: {'cosine_distance': []}, 1000: {'cosine_distance': []}, 5000: {'cosine_distance': []}, 20000: {'cosine_distance': []}}
        
        self.dist_classes = [DistanceResolver().resolve(x) for x in self.distance_types]


    def base_check(self, inputs, outputs, gts=None, extra_args={}):
        vals = self.measurable.compute_and_log(
            inputs, outputs, gts=gts, extra=extra_args
        )
        aggregate_ids = self.aggregate_measurable.compute_and_log(
            inputs, outputs, gts=gts, extra=extra_args
        )
        counts = self.count_measurable.compute_and_log(
            inputs, outputs, gts=gts, extra=extra_args
        )
        all_models = [x.compute_and_log(
            inputs, outputs, gts=gts, extra=extra_args
        ) for x in self.model_measurables]
        all_features = dict(zip(self.feature_names, 
            [x.compute_and_log(inputs, outputs, gts=gts, extra=extra_args) 
            for x in self.feature_measurables]))

        models = dict(zip(['model_' + x for x in self.model_names], [self.allowed_model_values[jdx][0] for jdx in range(len(self.model_names))]))
        for idx in range(len(aggregate_ids)):
            is_model_invalid = sum([all_models[jdx][idx] not in self.allowed_model_values[jdx] for jdx in range(len(self.allowed_model_values))])
            if is_model_invalid:
                continue

            if aggregate_ids[idx] not in self.item_counts:
                self.item_counts.update({aggregate_ids[idx]: 0})

            this_item_count_prev = self.item_counts[aggregate_ids[idx]]
            if this_item_count_prev == counts[idx]:
                continue
            self.item_counts[aggregate_ids[idx]] = counts[idx]
            this_item_count = self.item_counts[aggregate_ids[idx]]

            crossed_checkpoint_arr = (np.logical_xor((self.count_checkpoints > this_item_count), 
                (self.count_checkpoints > this_item_count_prev)))
            has_crossed_checkpoint = sum(crossed_checkpoint_arr)

            if has_crossed_checkpoint:

                if aggregate_ids[idx] not in self.vals_dictn[0]:
                    crossed_checkpoint = 0
                else:
                    crossed_checkpoint = self.count_checkpoints[np.where(crossed_checkpoint_arr == True)[0][-1]]

                this_val = extract_data_points_from_batch(vals, [idx])

                if aggregate_ids[idx] not in self.vals_dictn[crossed_checkpoint]:
                    self.vals_dictn[crossed_checkpoint].update({aggregate_ids[idx]: {}})
                self.vals_dictn[crossed_checkpoint][aggregate_ids[idx]].update({'val': this_val})
                self.vals_dictn[crossed_checkpoint][aggregate_ids[idx]].update({'visual_hover_text_' + self.aggregate_measurable.col_name(): aggregate_ids[idx]})
                for feat_name, feats in all_features.items():
                    this_feat = extract_data_points_from_batch(feats, [idx])
                    self.vals_dictn[crossed_checkpoint][aggregate_ids[idx]].update({'visual_label_' + feat_name: this_feat})

                selected_ids_to_sample = random.choices(list(self.vals_dictn[crossed_checkpoint].keys()), k=min(10, len(self.vals_dictn[crossed_checkpoint].keys())))
                for agg_i in selected_ids_to_sample:
                    if agg_i == aggregate_ids[idx]:
                        continue
                    this_distances = dict(
                        zip(
                            self.distance_types,
                            [
                                x.compute_distance(
                                    this_val, self.vals_dictn[crossed_checkpoint][agg_i]['val']
                                )
                                for x in self.dist_classes
                            ],
                        )
                    )

                    for distance_key in list(this_distances.keys()):

                        plot_name = distance_key
                        this_data = list(np.reshape(np.array(this_distances[distance_key]),-1))

                        self.log_handler.add_histogram(
                            plot_name,
                            this_data,
                            self.dashboard_name,
                            models = [models]*len(this_data),
                            file_name = str(crossed_checkpoint)
                        )

    
    def get_feats_for_clustering(self, count, allowed_model_values):
        if len(self.children) > 0:
            res = {}
            feats = {}
            for x in self.children:
                val = x.get_feats_for_clustering(count, allowed_model_values)
                res.update(val)
            return res
        else:
            if self.allowed_model_values == allowed_model_values:
                if count in self.vals_dictn:
                    return self.vals_dictn[count]
                else:
                    return {}
            else:
                return {}


import collections

from typing import List


class PopularityBias:
    """A class to estimate item popularity bias from a set of sessions and add predictions to it."""

    def __init__(self, sessions: List[List[str]], k: int = 10) -> None:
        """
        Initializes the PopularityBias object and estimates item popularity from sessions.

        Parameters
        ----------
        sessions
            A list of sessions, where each session is a list of item IDs.
        k
            The number of recommendations to make. Defaults to 10.
        """

        self.k = k
        self.all_popularity = []

        # Estimate item popularity from sessions
        pop_map = collections.defaultdict(lambda: 0)
        num_interactions = 0

        for session in sessions:
            for event in session:
                pop_map[event] += 1
                num_interactions += 1

        # Normalize item popularity
        self.pop_map = {k: v / num_interactions for k, v in pop_map.items()}

    def add_prediction(self, prediction: List[str]) -> None:
        """
        Adds a prediction to the all_popularity list.

        Parameters
        ----------
        prediction
            A list of item IDs for a single prediction.
        """

        # Get popularity score for each item in the prediction and add it
        pop_one_pred = [self.pop_map.get(event, 0.0) for event in prediction]
        self.all_popularity += pop_one_pred



"""
A setuptools based setup module.
"""
#!/usr/bin/env python

# Always prefer setuptools over distutils
from setuptools import setup, find_packages

# read the contents of your README file
from pathlib import Path

this_directory = Path(__file__).parent
long_description = (this_directory / "README.md").read_text()

setup(
    name="uptrain",
    version="0.0.10",
    description="UpTrain - ML Observability and Retraining Framework",
    long_description=long_description,
    long_description_content_type="text/markdown",
    # The project's main homepage.
    url="https://github.com/uptrain-ai/uptrain",
    # Author details
    maintainer="UpTrain AI Team",
    maintainer_email="uptrain.ai@gmail.com",
    # Choose your license
    license="Apache License 2.0",
    # See https://pypi.python.org/pypi?%3Aaction=list_classifiers
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Intended Audience :: Developers",
        "Topic :: Software Development :: Build Tools",
        # Pick your license as you wish (should match "license" above)
        "License :: OSI Approved :: Apache Software License",
        # Specify the Python versions you support here. In particular, ensure
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
    ],
    keywords="uptrain ai retraining ML observability",
    packages=find_packages(),
    install_requires=[
        "numpy<1.24",
        "umap-learn>=0.5.3",
        "pandas>=1.0.0",
        "plotly>=5.0.0",
        "pydantic>=1.9.0",
        "river<=0.14",
        "scikit_learn>=1.0.0",
        "streamlit>=1.0.0",
        "shap>=0.41.0",
        "json-fix>=0.5.0",
    ],
    tests_require=["pytest>=7.0", "torch", "imgaug", "gensim", "xgboost", "gensim"],
)



import numpy as np


class L2Distance:
    def compute_distance(self, base, ref) -> float:
        base = np.array(base)
        ref = np.array(ref)
        base_shape = list(base.shape)
        return np.sum(
            (base - ref) * (base - ref), axis=tuple(range(1, len(base_shape)))
        )



import numpy as np
from uptrain.core.classes.monitors import AbstractMonitor
from uptrain.constants import Monitor


class CustomMonitor(AbstractMonitor):
    monitor_type = Monitor.CUSTOM_MONITOR

    def base_init(self, fw, check):
        self.dashboard_name = check.get("dashboard_name", "custom_measure")
        self.check_func = check["check_func"]
        self.need_gt = check["need_gt"]

        initialize_func = check.get("initialize_func", None)
        if initialize_func is not None:
            initialize_func(self)

    def base_check(self, inputs, outputs, gts=None, extra_args={}):
        check_func = self.check_func
        return check_func(self, inputs, outputs, gts=gts, extra_args=extra_args)

    def need_ground_truth(self):
        return self.need_gt



import numpy as np
from copy import deepcopy

from uptrain.constants import Monitor, Statistic, Visual, MeasurableType
from uptrain.core.classes.monitors import (
    Accuracy,
    ConceptDrift,
    DataDrift,
    CustomMonitor,
    ModelBias,
    DataIntegrity,
    EdgeCase,
)
from uptrain.core.classes.statistics import (
    Distance,
    Convergence,
    Distribution,
)
from uptrain.core.classes.visuals import Umap, Tsne, Shap


class CheckManager:
    def __init__(self, framework, checks=[]):
        self.monitors_to_check = []
        self.statistics_to_check = []
        self.visuals_to_check = []
        self.fw = framework
        for check in checks:
            if check["type"] in Monitor:
                self.add_monitor(check)
            if check["type"] in Statistic:
                self.add_statistic(check)
            if check["type"] in Visual:
                self.add_visual(check)

    def add_monitor(self, check):
        if check["type"] == Monitor.EDGE_CASE:
            edge_case_manager = EdgeCase(self.fw, check)
            self.monitors_to_check.append(edge_case_manager)
        elif check["type"] == Monitor.ACCURACY:
            acc_manager = Accuracy(self.fw, check)
            self.monitors_to_check.append(acc_manager)
        elif check["type"] == Monitor.CONCEPT_DRIFT:
            drift_manager = ConceptDrift(self.fw, check)
            self.monitors_to_check.append(drift_manager)
        elif check["type"] == Monitor.DATA_DRIFT:
            if "measurable_args" in check:
                drift_managers = [DataDrift(self.fw, check)]
            else:
                drift_managers = []
                all_feats = self.fw.feat_name_list
                for feat in all_feats:
                    check_copy = deepcopy(check)
                    check_copy.update(
                        {
                            "measurable_args": {
                                "type": MeasurableType.INPUT_FEATURE,
                                "feature_name": feat,
                            }
                        }
                    )
                    drift_managers.append(DataDrift(self.fw,check_copy))
            self.monitors_to_check.extend(drift_managers)
        elif check["type"] == Monitor.POPULARITY_BIAS:
            bias_manager = ModelBias(self.fw, check)
            self.monitors_to_check.append(bias_manager)
        elif check["type"] == Monitor.CUSTOM_MONITOR:
            custom_monitor = CustomMonitor(self.fw, check)
            self.monitors_to_check.append(custom_monitor)
        elif check["type"] == Monitor.DATA_INTEGRITY:
            custom_monitor = DataIntegrity(self.fw, check)
            self.monitors_to_check.append(custom_monitor)
        else:
            raise Exception("Monitor type not Supported")

    def add_statistic(self, check):
        if check["type"] == Statistic.DISTANCE:
            custom_monitor = Distance(self.fw, check)
            self.statistics_to_check.append(custom_monitor)
        elif check["type"] == Statistic.DISTRIBUTION_STATS:
            custom_monitor = Distribution(self.fw, check)
            self.statistics_to_check.append(custom_monitor)
        elif check["type"] == Statistic.CONVERGENCE_STATS:
            custom_monitor = Convergence(self.fw, check)
            self.statistics_to_check.append(custom_monitor)
        else:
            raise Exception("Statistic type not Supported")

    def add_visual(self, check):
        if check["type"] == Visual.UMAP:
            custom_monitor = Umap(self.fw, check)
            self.visuals_to_check.append(custom_monitor)
        elif check["type"] == Visual.TSNE:
            custom_monitor = Tsne(self.fw, check)
            self.visuals_to_check.append(custom_monitor)
        elif check["type"] == Visual.SHAP:
            custom_monitor = Shap(self.fw, check)
            self.visuals_to_check.append(custom_monitor)
        else:
            raise Exception("Visual type not Supported")

    def check(self, inputs, outputs, gts=None, extra_args={}):
        for monitor in self.monitors_to_check:
            if monitor.need_ground_truth() == (gts[0] is not None):
                monitor.check(inputs, outputs, gts=gts, extra_args=extra_args)
        for stats in self.statistics_to_check:
            stats.check(inputs, outputs, gts=gts, extra_args=extra_args)
        for visuals in self.visuals_to_check:
            visuals.check(inputs, outputs, gts=gts, extra_args=extra_args)

    def is_data_interesting(self, inputs, outputs, gts=None, extra_args={}):
        is_interesting = []
        reasons = []
        for monitor in self.monitors_to_check:
            if monitor.need_ground_truth() == (gts[0] is not None):
                res = monitor.is_data_interesting(
                        inputs, outputs, gts=gts, extra_args=extra_args
                    )
                is_interesting.append(res[0])
                reasons.append(res[1])
        if len(reasons):
            final_reason = ["None"] * len(reasons[0])
            for reas in reasons:
                for jdx in range(len(reas)):
                    if not (reas[jdx] == "None"):
                        final_reason[jdx] = reas[jdx]
        else:
            final_reason = []
        return np.greater(np.sum(np.array(is_interesting), axis=0), 0), np.array(final_reason)



from uptrain.core.lib.helper_funcs import read_json


class AnnotationHelper:
    """
    Helper class to integrate different annotation methods

    ...

    Methods
    -------
    annotations_from_master_file(dataset):
        Reads gt from the master_file. Expects matching_key (defaulted to id) and master_file to be present in args
    """

    def __init__(self, args={}):
        self.args = args

    def annotations_from_master_file(self, dataset):
        """Reads gt annotation from a master_file using matching_key as the index
        Expects location of master_file in args
        Defaults matching_key to id if not present in args
        """

        matching_key = self.args.get("matching_key", "id")
        data = read_json(dataset)
        master_data = read_json(self.args["master_file"])
        all_master_keys = [x[matching_key] for x in master_data]
        for datapoint in data:
            datapoint.update(
                {
                    "gt": master_data[all_master_keys.index(datapoint[matching_key])][
                        "gt"
                    ]
                }
            )
        return data
